#!/usr/bin/env python
import pytest


@pytest.mark.asyncio
async def test_WhoHas(network_and_devices):
    # Write to an object and validate new value is correct
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        response = bacnet.whohas(
            "analogInput:0", destination="{}".format(test_device.properties.address)
        )
        assert response
        # response = bacnet.whohas("analogInput:0", global_broadcast=True)
        # assert response
        # response = bacnet.whohas("analogInput:0")
        # assert response
        # Can't work as I'm using different ports to have multiple devices using the same IP....
        # So neither local or global broadcast will give result here



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Lite is the base class to create a BACnet network
It uses provided args to register itself as a device in the network
and allow communication with other devices.

"""

import asyncio
import typing as t

# --- standard Python modules ---
import weakref

from bacpypes3 import __version__ as bacpypes_version
from bacpypes3.app import Application
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import ObjectIdentifier

from BAC0.core.app.asyncApp import BAC0Application
from BAC0.scripts.Base import Base

from ..core.devices.Device import RPDeviceConnected, RPMDeviceConnected
from ..core.devices.Points import Point
from ..core.devices.Trends import TrendLog
from ..core.devices.Virtuals import VirtualPoint
from ..core.functions.Alias import Alias
from ..core.functions.CoV import COVSubscription

# from ..core.functions.legacy.cov import CoV
# from ..core.functions.legacy.DeviceCommunicationControl import (
#    DeviceCommunicationControl,
# )
from ..core.functions.Discover import Discover
from ..core.functions.EventEnrollment import EventEnrollment
from ..core.functions.GetIPAddr import HostIP
from ..core.functions.Reinitialize import Reinitialize

# from ..core.functions.legacy.Reinitialize import Reinitialize
from ..core.functions.Schedule import Schedule
from ..core.functions.Text import TextMixin
from ..core.functions.TimeSync import TimeSync
from ..core.io.IOExceptions import (
    NoResponseFromController,
    NumerousPingFailures,
    Timeout,
    UnrecognizedService,
)
from ..core.io.Read import ReadProperty
from ..core.io.Simulate import Simulation
from ..core.io.Write import WriteProperty
from ..core.utils.lookfordependency import influxdb_if_available, rich_if_available

# from ..core.io.asynchronous.Write import WriteProperty
from ..core.utils.notes import note_and_log
from ..infos import __version__ as version

# --- this application's modules ---
from ..tasks.RecurringTask import RecurringTask
from ..tasks.TaskManager import Task

INFLUXDB, _ = influxdb_if_available()
if INFLUXDB:
    from ..db.influxdb import InfluxDB
RICH, rich = rich_if_available()
if RICH:
    from rich import pretty
    from rich.console import Console
    from rich.table import Table

    pretty.install()


# ------------------------------------------------------------------------------


@note_and_log
class Lite(
    Base,
    Discover,
    Alias,
    EventEnrollment,
    ReadProperty,
    WriteProperty,
    Simulation,
    TimeSync,
    Reinitialize,
    # DeviceCommunicationControl,
    COVSubscription,
    Schedule,
    # Calendar,
    TextMixin,
):
    """
    Build a BACnet application to accept read and write requests.
    [Basic Whois/IAm functions are implemented in parent BasicScript class.]
    Once created, execute a whois() to build a list of available controllers.
    Initialization requires information on the local device.

    :param ip='127.0.0.1': Address must be in the same subnet as the BACnet network
        [BBMD and Foreign Device - not supported]

    """

    def __init__(
        self,
        ip: t.Optional[str] = None,
        port: t.Optional[int] = None,
        mask: t.Optional[int] = None,
        bbmdAddress=None,
        bbmdTTL: int = 0,
        bdtable=None,
        ping: bool = True,
        ping_delay: int = 300,
        db_params: t.Optional[t.Dict[str, t.Any]] = None,
        **params,
    ) -> None:
        self._initialized = False
        self.log(
            f"Starting Asynchronous BAC0 version {version} ({self.__module__.split('.')[-1]})",
            level="info",
        )
        self.log(f"Using bacpypes3 version {bacpypes_version}", level="info")
        self.log("Use BAC0.log_level to adjust verbosity of the app.", level="info")
        self.log(
            "Ex. BAC0.log_level('silence') or BAC0.log_level('error')", level="info"
        )

        self.log("Configurating app", level="debug")
        self._registered_devices = weakref.WeakValueDictionary()

        # Ping task will deal with all registered device and disconnect them if they do not respond.

        self._ping_task = RecurringTask(
            self.ping_registered_devices,
            delay=ping_delay,
            name="Ping Registered Devices Task",
        )
        if ping:
            self._ping_task.start()

        self._cleanup_task = RecurringTask(
            Task.clean_tasklist, delay=60, name="Cleanup Tasks List"
        )
        self._cleanup_task.start()

        if ip is None:
            host = HostIP(port)
            mask = host.mask
            ip_addr = host.address
        else:
            try:
                ip, subnet_mask_and_port = ip.split("/")
                try:
                    mask_s, port_s = subnet_mask_and_port.split(":")
                    mask = int(mask_s)
                    port = int(port_s)
                except ValueError:
                    mask = int(subnet_mask_and_port)
            except ValueError:
                ip = ip

            if not mask:
                mask = 24
            if not port:
                port = 47808
            ip_addr = Address(f"{ip}/{mask}:{port}")
        self._log.info(
            f"Using ip : {ip_addr}/{mask} on port {ip_addr.addrPort} | broadcast : {ip_addr.addrBroadcastTuple[0]}"
        )

        Base.__init__(
            self,
            localIPAddr=ip_addr,
            bbmdAddress=bbmdAddress,
            bbmdTTL=bbmdTTL,
            bdtable=bdtable,
            **params,
        )
        self.log(f"Device instance (id) : {self.Boid}", level="info")
        self.bokehserver = False
        self._points_to_trend = weakref.WeakValueDictionary()

        # Activate InfluxDB if params are available
        if db_params and INFLUXDB:
            try:
                self.database = (
                    InfluxDB(db_params)
                    if db_params["name"].lower() == "influxdb"
                    else None
                )
                asyncio.create_task(
                    asyncio.wait_for(self.database._health(), timeout=5)
                )
            except TimeoutError:
                self._log.error(
                    "Unable to connect to InfluxDB. Please validate parameters"
                )
        if self.database:
            write_interval = db_params.get("write_interval", 60)
            self.create_save_to_influxdb_task(delay=write_interval)

        # Announce yourself

        self.i_am()

    def i_am(self):
        loop = asyncio.get_running_loop()
        loop.create_task(self._i_am())

    async def _i_am(self) -> None:
        while self.this_application.app is None or not asyncio.iscoroutinefunction(
            self.this_application.app.i_am
        ):
            await asyncio.sleep(0.01)
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        _res = await self.this_application.app.i_am()
        self._initialized = True

    def create_save_to_influxdb_task(self, delay: int = 60) -> None:
        self._write_to_db = RecurringTask(
            self.save_registered_devices_to_db,
            delay=delay,
            name="Write to InfluxDB Task",
        )
        self._write_to_db.start()

    async def save_registered_devices_to_db(self, delay: int = 60) -> None:
        if len(self.registered_devices) > 0:
            for each in self.registered_devices:
                try:
                    await self.database.write_points_lastvalue_to_db(each.points)
                except Exception as error:
                    self._log.error(
                        f"Error writing points of {each} to InfluxDB : {error}. Stopping task."
                    )
                    await self._write_to_db.stop()
                    self.log(
                        "Write to InfluxDB Task stopped. Restarting", level="warning"
                    )
                    self.create_save_to_influxdb_task(delay=delay)

    def register_device(
        self, device: t.Union[RPDeviceConnected, RPMDeviceConnected]
    ) -> None:
        oid = id(device)
        self._registered_devices[oid] = device

    async def ping_registered_devices(self) -> None:
        """
        Registered device on a network (self) are kept in a list (registered_devices).
        This function will allow pinging thoses device regularly to monitor them. In case
        of disconnected devices, we will disconnect the device (which will save it). Then
        we'll ping again until reconnection, where the device will be bring back online.

        To permanently disconnect a device, an explicit device.disconnect(unregister=True [default value])
        will be needed. This way, the device won't be in the registered_devices list and
        BAC0 won't try to ping it.
        """
        for each in self.registered_devices:
            if isinstance(each, RPDeviceConnected) or isinstance(
                each, RPMDeviceConnected
            ):
                try:
                    self._log.debug(
                        f"Ping {each.properties.name}|{each.properties.address}"
                    )
                    await each.ping()
                    if each.properties.ping_failures > 3:
                        raise NumerousPingFailures

                except NumerousPingFailures:
                    self._log.warning(
                        "{}|{} is offline, disconnecting it.".format(
                            each.properties.name, each.properties.address
                        )
                    )
                    await each._disconnect(unregister=False)

            else:
                device_id = each.properties.device_id
                addr = each.properties.address
                name = await self.read(f"{addr} device {device_id} objectName")
                if name == each.properties.name:
                    each.properties.ping_failures = 0
                    self._log.info(
                        "{}|{} is back online, reconnecting.".format(
                            each.properties.name, each.properties.address
                        )
                    )
                    each.connect(network=self)
                    each.poll(delay=each.properties.pollDelay)

    @property
    def registered_devices(self):
        """
        Devices that have been created using BAC0.device(args)
        """
        return list(self._registered_devices.values())

    def unregister_device(self, device):
        """
        Remove from the registered list
        """
        oid = id(device)
        try:
            del self._registered_devices[oid]
        except KeyError:
            pass

    def add_trend(self, point_to_trend: t.Union[Point, TrendLog, VirtualPoint]) -> None:
        """
        Add point to the list of histories that will be handled by Bokeh

        Argument provided must be of type Point or TrendLog
        ex. bacnet.add_trend(controller['point_name'])
        """
        if (
            isinstance(point_to_trend, Point)
            or isinstance(point_to_trend, TrendLog)
            or isinstance(point_to_trend, VirtualPoint)
        ):
            oid = id(point_to_trend)
            self._points_to_trend[oid] = point_to_trend
        else:
            raise TypeError("Please provide point containing history")

    def remove_trend(
        self, point_to_remove: t.Union[Point, TrendLog, VirtualPoint]
    ) -> None:
        """
        Remove point from the list of histories that will be handled by Bokeh

        Argument provided must be of type Point or TrendLog
        ex. bacnet.remove_trend(controller['point_name'])
        """
        if (
            isinstance(point_to_remove, Point)
            or isinstance(point_to_remove, TrendLog)
            or isinstance(point_to_remove, VirtualPoint)
        ):
            oid = id(point_to_remove)
        else:
            raise TypeError("Please provide point or trendLog containing history")
        if oid in self._points_to_trend.keys():
            del self._points_to_trend[oid]

    @property
    async def devices(self):
        await self._devices(_return_list=False)

    async def _devices(
        self, _return_list: bool = False
    ) -> t.List[t.Tuple[str, str, str, int]]:
        """
        This property will create a good looking table of all the discovered devices
        seen on the network.

        For that, some requests will be sent over the network to look for name,
        manufacturer, etc and in big network, this could be a long process.
        """

        lst = []
        if self.discoveredDevices is not None:
            for k, v in self.discoveredDevices.items():
                objid, device_address, network_number, vendor_id, vendor_name = (
                    v["object_instance"],
                    v["address"],
                    v["network_number"],
                    v["vendor_id"],
                    v["vendor_name"],
                )
                devId = objid[1]
                try:
                    deviceName, vendorName = await self.readMultiple(
                        f"{device_address} device {devId} objectName vendorName"
                    )
                except (UnrecognizedService, ValueError):
                    self._log.warning(
                        f"Unrecognized service for {devId} | {device_address}"
                    )
                    try:
                        deviceName = await self.read(
                            f"{device_address} device {devId} objectName"
                        )
                        vendorName = await self.read(
                            f"{device_address} device {devId} vendorName"
                        )
                    except NoResponseFromController:
                        self.log(f"No response from {k}", level="warning")
                        continue
                except (NoResponseFromController, Timeout):
                    self.log(f"No response from {k}", level="warning")
                    continue
                lst.append(
                    (deviceName, vendorName, devId, device_address, network_number)
                )
            if RICH:
                console = Console()
                table = Table(show_header=True, header_style="bold magenta")
                table.add_column("Network_number")
                table.add_column("Device Name")
                table.add_column("Address")
                table.add_column("Device Instance")
                table.add_column("Vendor Name")
                for each in lst:
                    deviceName, vendorName, devId, device_address, network_number = each
                    table.add_row(
                        f"{network_number}",
                        f"{deviceName}",
                        f"{device_address}",
                        f"{devId}",
                        f"{vendorName}",
                    )
                console.print(table)
        if _return_list:
            return lst  # type: ignore[return-value]

    @property
    def trends(self) -> t.List[t.Any]:
        """
        This will present a list of all registered trends used by Bokeh Server
        """
        return list(self._points_to_trend.values())

    @property
    def tasks(self) -> t.List[Task]:
        """
        This will present a list of all registered tasks
        """
        return Task.tasks

    def disconnect(self) -> None:
        asyncio.create_task(self._disconnect())

    async def _disconnect(self) -> None:
        self.log("Disconnecting", level="debug")
        for each in self.registered_devices:
            await each._disconnect()
        await super()._disconnect()
        self._initialized = False

    def __repr__(self) -> str:
        return f"Bacnet Network using ip {self.localIPAddr} with device id {self.Boid}"

    def __getitem__(self, boid_or_localobject: t.Union[str, ObjectIdentifier, tuple]):
        """
        Retrieve an item from the application by its name or identifier.

        Args:
            boid_or_localobject (Union[str, ObjectIdentifier]): The name (as a string) or identifier (as an ObjectIdentifier) of the object to retrieve.

        Returns:
            Union[YourObjectType, None]: The object corresponding to the given name or identifier, or a registered device if the identifier matches a device ID. Returns None if the object or device is not found.

        Raises:
            KeyError: If the object or device is not found and logging is not enabled.
        """
        if isinstance(boid_or_localobject, str):
            item = self.this_application.app.objectName[boid_or_localobject]
        elif isinstance(boid_or_localobject, ObjectIdentifier):
            item = self.this_application.app.objectIdentifier[boid_or_localobject]
        elif isinstance(boid_or_localobject, tuple):
            _objId = ObjectIdentifier(boid_or_localobject)
            item = self.this_application.app.objectIdentifier[_objId]
        if item is None:
            for device in self._registered_devices:
                if str(device.properties.device_id) == str(boid_or_localobject):
                    return device
            self.log(f"{boid_or_localobject} not found", level="error")
        else:
            return item

    async def __aenter__(self):
        while not self._initialized:
            await asyncio.sleep(0.1)
        self._log.info(
            f"{self.localObjName}|{self.Boid} connected. Entering context manager."
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._disconnect()
        while self._initialized:
            await asyncio.sleep(0.1)
        self._log.info(
            f"{self.localObjName}|{self.Boid} disconnected. Exiting context manager."
        )

    def get_device_by_id(self, id):
        for each in self.registered_devices:
            if each.properties.device_id == id:
                return each
        self._log.error(f"Device {id} not found")
        raise ValueError("Device not found")

    def cov(
        self,
        address: str = None,
        objectID: t.Tuple[str, int] = None,
        lifetime: int = 900,
        confirmed: bool = False,
        callback: t.Optional[
            t.Union[t.Callable[[str, t.Any], None], t.Awaitable[None]]
        ] = None,
    ):
        """
        Subscribe to COV notification for a given address and objectID
        If callback is provided, it will be called with the value of the COV notification
        So the function must be built to accept two arguments : property_identifier and property_value
        """
        cov_task = COVSubscription(
            address=address,
            objectID=objectID,
            confirmed=confirmed,
            lifetime=lifetime,
            callback=callback,
            BAC0App=self,
        )
        Base._running_cov_tasks[cov_task.process_identifier] = cov_task
        cov_task.task = asyncio.create_task(cov_task.run())
        self.log(
            f"COV subscription for {address}|{objectID} with id {cov_task.process_identifier} started",
            level="info",
        )

    def cancel_cov(self, task_id: int):
        self.log(f"Canceling COV subscription id {task_id}", level="info")
        process_identifer = task_id

        if process_identifer not in Base._running_cov_tasks:
            self.log(f"Task {process_identifer} not found", level="warning")
            return
        cov_subscription = Base._running_cov_tasks.pop(process_identifer)
        cov_subscription.stop()
        # await cov_subscription.task

    @property
    def cov_tasks(self):
        return Base._running_cov_tasks



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Write.py - creation of WriteProperty requests

    Used while defining an app
    Example::

        class BasicScript(WhoisIAm, WriteProperty)

    Class::

        WriteProperty()
            def write()


"""
import re

from bacpypes3.apdu import ErrorRejectAbortNack
from bacpypes3.app import Application
from bacpypes3.basetypes import PropertyIdentifier

# --- 3rd party modules ---
from bacpypes3.debugging import ModuleLogger
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import Null, ObjectIdentifier
from bacpypes3.errors import PropertyError

from BAC0.tasks.DoOnce import DoOnce

from ..app.asyncApp import BAC0Application
from ..utils.notes import note_and_log

# --- this application's modules ---
from .IOExceptions import (
    ApplicationNotStarted,
    NoResponseFromController,
    WritePropertyException,
)

# ------------------------------------------------------------------------------

# some debugging
_debug = 0
_LOG = ModuleLogger(globals())
WRITE_REGEX = r"(?P<address>\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(?::\d+)?\b|(\b\d+:\d+\b)) (?P<objId>(@obj_)?[-\w:]*[: ]*\d*) (?P<propId>(@prop_)?\w*(-\w*)?)[ ]?(?P<value>-?[\w\d]*\.?\d*)?[ ]?(?P<indx>-|\d*)?[ ]?(?P<priority>(1[0-6]|[0-9]))?"
write_pattern = re.compile(WRITE_REGEX)


@note_and_log
class WriteProperty:
    """
    Defines BACnet Write functions: WriteProperty [WritePropertyMultiple not supported]

    """

    def write(self, args, vendor_id=0, timeout=10):
        # asyncio.create_task(
        #    self._write(args=args, vendor_id=vendor_id, timeout=timeout)
        # )
        # loop = asyncio.get_event_loop()
        # loop.run_until_complete(self._write(args=args, vendor_id=vendor_id, timeout=timeout))
        write_task = DoOnce((self._write, [args, vendor_id, timeout]))
        write_task.start()

    async def _write(self, args, vendor_id=0, timeout=10):
        """Build a WriteProperty request, wait for an answer, and return status [True if ok, False if not].

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] - [ <priority> ]
        :returns: return status [True if ok, False if not]

        *Example*::

            import BAC0
            bacnet = BAC0.lite()
            bacnet.write('2:5 analogValue 1 presentValue 100 - 8')

        Direct the controller at (Network 2, address 5) to write 100 to the presentValues of
        its analogValue 1 (AV:1) at priority 8
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log_title("Write property", args)

        (
            device_address,
            object_identifier,
            property_identifier,
            value,
            property_array_index,
            priority,
        ) = request = self.build_wp_request(args)
        # print(request)

        try:
            response = await _app.write_property(
                device_address,
                object_identifier,
                property_identifier,
                value,
                property_array_index,
                priority,
            )
            if response == "-no property type-" or response == "-no object class-":
                self.log(
                    f"exception: {response!r}, you probably need to define proprietary objects and property for this vendor (or import the specific module)",
                    level="error",
                )
            return response

        except ErrorRejectAbortNack as err:
            self.log(f"exception: {err!r}", level="error")
            raise NoResponseFromController(f"APDU Abort Reason : {response}")

        except ValueError as err:
            self.log(f"exception: {err!r}", level="error")
            raise ValueError(f"Invalid value for property : {err} | {response}")

        except (WritePropertyException, PropertyError) as error:
            # construction error
            self.log(f"exception: {error!r}", level="error")

        except TypeError as error:
            # construction error
            self.log(f"exception: {error!r} | Requests = {request}", level="error")

    @classmethod
    def _parse_wp_args(cls, args):
        """
        Utility to parse the string of the request.
        Supports @obj_ and @prop_ syntax for objest type and property id, useful with proprietary objects and properties.
        """
        global write_pattern
        match = write_pattern.search(args)
        try:
            address = match.group("address")
            objId = match.group("objId")
            prop_id = match.group("propId")
            value = match.group("value")
            indx = match.group("indx")
            priority = match.group("priority")
        except AttributeError:
            raise ValueError(f"Invalid request | {args}")
        if ":" in objId:
            obj_type, obj_inst = objId.split(":")
        else:
            obj_type, obj_inst = objId.split(" ")
        if obj_type.isdigit():
            obj_type = int(obj_type)
        elif "@obj_" in obj_type:
            obj_type = int(obj_type.split("_")[1])
        obj_inst = int(obj_inst)
        if "@prop_" in prop_id:
            prop_id = prop_id.split("_")[1]
        if prop_id.isdigit():
            prop_id = int(prop_id)

        value = Null(()) if value == "null" else value
        indx = None if indx == "-" or indx is None or indx == "" else int(indx)
        priority = None if priority is None else int(priority)

        return (address, obj_type, obj_inst, prop_id, value, priority, indx)

    def build_wp_request(self, args, vendor_id=0):
        vendor_id = vendor_id
        (
            address,
            obj_type,
            obj_inst,
            prop_id,
            value,
            priority,
            indx,
        ) = WriteProperty._parse_wp_args(args)

        object_identifier = ObjectIdentifier((obj_type, obj_inst))
        property_identifier = PropertyIdentifier(prop_id)
        device_address = Address(address)

        request = (
            device_address,
            object_identifier,
            property_identifier,
            value,
            indx,
            priority,
        )

        self.log_subtitle("Creating Request")
        self.log(
            f"{'indx':<20} {'priority':<20} {'datatype':<20} {'value':<20}",
            level="debug",
        )

        self.log(f"{'REQUEST':<20} {request}", level="debug")
        return request


'''
    def writeMultiple(self, addr=None, args=None, vendor_id=0, timeout=10):
        """Build a WritePropertyMultiple request, wait for an answer

        :param addr: destination of request (ex. '2:3' or '192.168.1.2')
        :param args: list of String with <type> <inst> <prop> <value> [ <indx> ] - [ <priority> ]
        :param vendor_id: Mandatory for registered proprietary object and properties
        :param timeout: used by IOCB to discard request if timeout reached
        :returns: return status [True if ok, False if not]

        *Example*::

            import BAC0
            bacnet = BAC0.lite()
            r = ['analogValue 1 presentValue 100','analogValue 2 presentValue 100','analogValue 3 presentValue 100 - 8','@obj_142 1 @prop_1042 True']
            bacnet.writeMultiple(addr='2:5',args=r,vendor_id=842)
            # or
            # bacnet.writeMultiple('2:5',r)

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        self.log_title("Write property multiple", args)

        try:
            # build a WritePropertyMultiple request
            iocb = IOCB(self.build_wpm_request(args, vendor_id=vendor_id, addr=addr))
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self.log(f"{'iocb':<20} {iocb!r}", level='debug')

        except WritePropertyException as error:
            # construction error
            self._log.exception(f"exception: {error!r}")

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self.log("Not an ack, see debug for more infos.", level='warning')
                self.log(f"Not an ack. | APDU : {apdu} / {type(apdu)}", level='debug')
                return

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController(f"APDU Abort Reason : {reason}")

    def build_wpm_request(self, args, vendor_id=0, addr=None):
        if not addr:
            raise ValueError("Please provide addr")

        address = Address(addr)

        self.log_subtitle("Creating Write Multiple Request")
        self.log(f"{'indx':<20} {'priority':<20} {'datatype':<20} {'value':<20}", level='debug')

        was = []
        listOfObjects = []
        for each in args:
            property_values = []
            if isinstance(each, str):
                (
                    object_identifier,
                    property_identifier,
                    value,
                    priority,
                    indx,
                ) = self._parse_wp_args(each)
            elif isinstance(each, tuple):
                # Supported but not really the best choice as the parser will
                # catch some edge cases for you
                object_identifier, property_identifier, value, priority, indx = each
            else:
                raise ValueError("Wrong encoding of request")

            existingObject = next(
                (obj for obj in was if obj.objectIdentifier == (obj_type, obj_inst)),
                None,
            )

            if existingObject is None:
                property_values.append(
                    PropertyValue(
                        propertyIdentifier=prop_id,
                        propertyArrayIndex=indx,
                        value=value,
                        priority=priority,
                    )
                )

                was.append(
                    WriteAccessSpecification(
                        objectIdentifier=(obj_type, obj_inst),
                        listOfProperties=property_values,
                    )
                )
            else:
                existingObject.listOfProperties.append(
                    PropertyValue(
                        propertyIdentifier=prop_id,
                        propertyArrayIndex=indx,
                        value=value,
                        priority=priority,
                    )
                )

            datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)

            self._log.debug(
                f"{indx!r:<20} {priority!r:<20} {datatype!r:<20} {value!r:<20}"
            )

        # build a request
        request = WritePropertyMultipleRequest(listOfWriteAccessSpecs=was)
        request.pduDestination = Address(addr)

        self.log(f"{'REQUEST':<20} {request}", level='debug')
        return request
'''



import asyncio
import typing as t

from bacpypes3.app import Application
from bacpypes3.pdu import Address, LocalBroadcast
from bacpypes3.primitivedata import ObjectIdentifier

from BAC0.core.app.asyncApp import BAC0Application

from ...core.utils.notes import note_and_log


@note_and_log
class Discover:
    """
    Main function to explore the network and find devices.
    """

    @property
    def known_network_numbers(self) -> t.Set[int]:
        return self.this_application._learnedNetworks

    def discover(
        self,
        networks: t.Union[str, t.List[int], int] = "known",
        limits: t.Tuple[int, int] = (0, 4194303),
        global_broadcast: bool = False,
        reset: bool = False,
    ) -> None:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        asyncio.create_task(
            self._discover(
                networks=networks,
                limits=limits,
                global_broadcast=global_broadcast,
                reset=reset,
            )
        )

    async def _discover(
        self,
        networks: t.Union[str, t.List[int], int] = "known",
        limits: t.Tuple[int, int] = (0, 4194303),
        global_broadcast: bool = False,
        timeout: int = 3,
        reset: bool = False,
    ) -> None:
        """
        Discover is meant to be the function used to explore the network when we
        connect.
        It will trigger whois request using the different options provided with
        parameters.

        By default, a local broadcast will be used. This is required as in big
        BACnet network, global broadcast can lead to network flood and loss of data.

        If not parameters are given, BAC0 will try to :

            * Find the network on which it is
            * Find routers for other networks (accessible via local broadcast)
            * Detect "known networks"
            * Use the list of known networks and create whois request to find all devices on those networks

        This should be sufficient for most cases.

        Once discovery is done, user may access the list of "discovered devices" using ::

            bacnet.discoveredDevices

        :param networks (list, integer) : A simple integer or a list of integer
            representing the network numbers used to issue whois request.

        :param limits (tuple) : tuple made of 2 integer, the low limit and the high
            limit. Those are the device instances used in the creation of the
            whois request. Min : 0 ; Max : 4194303

        :param global_broadcast (boolean) : If set to true, a global broadcast
            will be used for the whois. Use with care.
        """
        if reset:
            self.discoveredDevices = {}
        found = []

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        _networks = _this_application._learnedNetworks

        deviceInstanceRangeLowLimit, deviceInstanceRangeHighLimit = limits
        # Try to find on which network we are
        _this_network = await self.what_is_network_number()
        if _this_network:
            _networks.add(_this_network)
        # Try to find local routers...
        _other_networks = await self.whois_router_to_network()
        if _other_networks == []:
            _other_networks = await self.whois_router_to_network(global_broadcast=True)
        for each in _other_networks:
            network_adapter, _iamrtn = each
            _networks.update(_iamrtn.iartnNetworkList)
        for net in _networks:
            _this_application._learnedNetworks.add(net)
        self.log(f"Found those networks : {self.known_network_numbers}", level="info")

        if networks:
            if isinstance(networks, list):
                # we'll make multiple whois...
                for network in networks:
                    if network < 65535:
                        _networks.add(network)
            elif networks == "known":
                _networks = self.known_network_numbers.copy()
            else:
                if isinstance(networks, int) and networks < 65535:
                    _networks.add(networks)

        if _networks and not global_broadcast:
            for each_network in _networks:
                self.log(f"Discovering network {each_network}", level="info")
                _res = await self.this_application.app.who_is(
                    low_limit=deviceInstanceRangeLowLimit,
                    high_limit=deviceInstanceRangeHighLimit,
                    address=Address(f"{each_network}:*"),
                    timeout=timeout,
                )
                for each in _res:
                    found.append((each, each_network))

        else:
            msg = (
                "Global braodacast required"
                if global_broadcast
                else "No BACnet network found"
            )
            self.log(
                f"{msg}, attempting a simple whois using provided device instances limits ({deviceInstanceRangeLowLimit} - {deviceInstanceRangeHighLimit})",
                level="info",
            )
            if global_broadcast is True:
                self._log.warning(
                    "Issuing a global Broadcast whois can create network flood. Use with care."
                )
            else:
                self.log("Issuing a local broadcast whois request.", level="info")
            _address = None if global_broadcast is True else LocalBroadcast()
            _res = await _app.who_is(
                low_limit=deviceInstanceRangeLowLimit,
                high_limit=deviceInstanceRangeHighLimit,
                address=_address,
                timeout=timeout,
            )
            for each in _res:
                found.append((each, _this_network))

        for iam_request, network_number in found:
            self.log(
                f"Found device {iam_request} on network {network_number}", level="debug"
            )
            if not self.discoveredDevices:
                self.discoveredDevices = {}  # we can add device as we found some...
            device_address: Address = iam_request.pduSource
            objid: ObjectIdentifier = iam_request.iAmDeviceIdentifier
            self.log(str(objid), level="debug")
            self.log(f"Discovered : {self.discoveredDevices.keys()}", level="debug")
            key = str(objid)
            if key in self.discoveredDevices:
                self._log.debug(
                    f"{objid} already in discovered devices. Adding network number {network_number} to the list."
                )
                self.discoveredDevices[key]["network_number"].add(network_number)
            else:
                self._log.debug(
                    f"Adding {objid} to discovered devices in network {network_number}."
                )
                self.discoveredDevices[key] = {
                    "object_instance": objid,
                    "address": device_address,
                    "network_number": {network_number},
                    "vendor_id": iam_request.vendorID,
                    "vendor_name": "unknown",
                }

        self.log(
            f"Discovery done. Found {len(self.discoveredDevices) if self.discoveredDevices else 0} devices on {len(_networks) if _networks else 0} BACnet networks.",
            level="info",
        )



import asyncio
from signal import SIGINT, SIGTERM, signal


def run(main_task, bacnet):
    global loop

    def handler(sig, sig2):
        if bacnet is not None:
            bacnet._log.info(f"Got signal: {sig!s}, shutting down.")
            bacnet.disconnect()
        loop.stop()

    try:
        loop = (
            asyncio.get_running_loop()
        )  # we also could have chosen get_event_loop(), and if we did we would not have needed to set_event_loop() and create a new event loop
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    # for sig in (SIGINT, SIGTERM):
    #    loop.add_signal_handler(sig, handler)
    signal(SIGINT, handler)
    signal(SIGTERM, handler)

    loop.create_task(main_task())
    loop.run_forever()
    tasks = asyncio.all_tasks(loop=loop)
    for t in tasks:
        t.cancel()
    group = asyncio.gather(*tasks, return_exceptions=True)
    loop.run_until_complete(group)
    loop.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
sql.py -
"""

import os.path

# --- standard Python modules ---
import pickle

# --- 3rd party modules ---
import aiosqlite

from ..core.io.IOExceptions import (
    DataError,
    NoResponseFromController,
    RemovedPointException,
)
from ..core.utils.lookfordependency import pandas_if_available

_PANDAS, pd, sql, Timestamp = pandas_if_available()
# --- this application's modules ---

# ------------------------------------------------------------------------------


class SQLMixin(object):
    """
    Use SQL to persist a device's contents.  By saving the device contents to an SQL
    database, you can work with the device's data while offline, or while the device
    is not available.
    """

    async def _read_from_sql(self, request, db_name):
        """
        Using the contextlib, I hope to close the connection to database when
        not in use
        """
        async with aiosqlite.connect(f"{db_name}.db") as con:
            async with con.execute(request) as cursor:
                rows = await cursor.fetchall()
                columns = [description[0] for description in cursor.description]
                return pd.DataFrame(rows, columns=columns)

    def dev_properties_df(self):
        dic = self.properties.asdict.copy()
        dic.pop("network", None)
        dic["objects_list"] = []
        dic.pop("pss", None)
        return dic

    def points_properties_df(self):
        """
        Return a dictionary of point/point_properties in preparation for storage in SQL.
        """
        pprops = {}
        for each in self.points:
            p = each.properties.asdict.copy()
            p.pop("device", None)
            p.pop("network", None)
            p.pop("simulated", None)
            p.pop("overridden", None)
            pprops[str(each.properties.name)] = p

        return pd.DataFrame(pprops)

    def backup_histories_df(self, resampling="1s"):
        """
        Build a dataframe of the point histories
        By default, dataframe will be resampled for 1sec intervals,
        NaN will be forward filled then backward filled. This way, no
        NaN values will remains and analytics will be easier.

        Please note that this can be disabled using resampling=False

        In the process of building the dataframe, analog values are
        resampled using the mean() function. So we have intermediate
        results between to records.

        For binary values, we'll use .last() so we won't get a 0.5 value
        which means nothing in this context.

        If saving a DB that already exists, previous resampling will survive
        the merge of old data and new data.
        """
        if not _PANDAS:
            self.log("Pandas is required to create dataframe.", level="error")
            return
        backup = {}
        if isinstance(resampling, str):
            resampling_needed = True
            resampling_freq = resampling
        elif resampling in [0, False]:
            resampling_needed = False

        def extract_value_and_string(val):
            if isinstance(val, str):
                if ":" in val:
                    _v, _s = val.split(":")
                    return (int(_v), _s)
                elif val == "active":
                    val = 1
                elif val == "inactive":
                    val = 0
            return (int(val), "unknown")

        # print(resampling, resampling_freq, resampling_needed)
        for point in self.points:
            _name = str(point.properties.name)
            try:
                if (
                    "binary" in point.properties.type
                    or "multi" in point.properties.type
                ):
                    backup[f"{_name}_str"] = (
                        point.history.apply(lambda x: extract_value_and_string(x)[1])
                        .resample(resampling_freq)
                        .last()
                    )
                    backup[_name] = (
                        point.history.apply(lambda x: extract_value_and_string(x)[0])
                        .resample(resampling_freq)
                        .last()
                    )
                elif resampling_needed and "analog" in point.properties.type:
                    backup[_name] = point.history.resample(resampling_freq).mean()
                else:
                    # backup[point.properties.name] = point.history.resample(
                    #    resampling_freq
                    # ).last()
                    continue

            except Exception as error:
                try:
                    self.log(
                        f"{self.properties.name} ({self.properties.device.properties.address}) | Error in resampling {point.properties.name} | {error} (probably not enough points)",
                        level="error",
                    )
                except AttributeError as error:
                    raise DataError(
                        f"Cannot save, missing required information : {error}"
                    )
                if (
                    "binary" in point.properties.type
                    or "multi" in point.properties.type
                ):
                    backup[f"{_name}.str"] = (
                        point.history.apply(lambda x: extract_value_and_string(x)[1])
                        .resample(resampling_freq)
                        .last()
                    )
                    backup[f"{_name}.val"] = (
                        point.history.apply(lambda x: extract_value_and_string(x)[0])
                        .resample(resampling_freq)
                        .last()
                    )
                    backup[_name] = point.history.resample(resampling_freq).last()
                elif "analog" in point.properties.type:
                    backup[_name] = point.history.resample(resampling_freq).mean()
                else:
                    # backup[point.properties.name] = point.history
                    continue

        df = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in backup.items()]))
        if resampling_needed:
            return df.resample(resampling_freq).last().ffill().bfill()
        else:
            return df

    async def save(self, filename=None, resampling=None):
        """
        Save the point histories to sqlite3 database.
        Save the device object properties to a pickle file so the device can be reloaded.

        Resampling : valid Pandas resampling frequency. If 0 or False, dataframe will not be resampled on save.
        """
        if not _PANDAS:
            self.log("Pandas is required to save to SQLite.", level="error")
            return

        if filename:
            if ".db" in filename:
                filename = filename.split(".")[0]
            self.properties.db_name = filename
        else:
            self.properties.db_name = f"Device_{self.properties.device_id}"

        if resampling is None:
            resampling = self.properties.save_resampling

        # Does file exist? If so, append data

        def _df_to_backup():
            try:
                return self.backup_histories_df(resampling=resampling)
            except (DataError, NoResponseFromController):
                self.log("Impossible to save right now, error in data", level="error")
                return pd.DataFrame()

        if os.path.isfile(f"{self.properties.db_name}.db"):
            try:
                his = await self._read_from_sql(
                    'select * from "history"', self.properties.db_name
                )
                his.index = his["index"].apply(Timestamp)
                last = his.index[-1]
                df_to_backup = _df_to_backup()[last:]
            except Exception:
                df_to_backup = _df_to_backup()

        else:
            self.log("Creating a new backup database", level="debug")
            df_to_backup = _df_to_backup()

        if df_to_backup is None:
            return
        # DataFrames that will be saved to SQL
        async with aiosqlite.connect(f"{self.properties.db_name}.db") as con:
            try:
                async with con.execute("SELECT * FROM history") as cursor:
                    data = await cursor.fetchall()
                    columns = [description[0] for description in cursor.description]
                    data = pd.DataFrame(data, columns=columns)
                    df = pd.concat([data, df_to_backup], sort=True)
                    sql.to_sql(
                        df_to_backup,
                        name="history",
                        con=con,
                        index_label="index",
                        index=True,
                        if_exists="append",
                    )
            except Exception:
                # df = df_to_backup
                self._log.error("Error saving to SQL database")

            # asyncio.run(
            #    None, df_to_backup.to_sql, "history", con, None, "append", True, "index"
            # )

        # Saving other properties to a pickle file...
        prop_backup = {"device": self.dev_properties_df()}
        prop_backup["points"] = self.points_properties_df()
        try:
            with open(f"{self.properties.db_name}.bin", "wb") as file:
                pickle.dump(prop_backup, file)
            if self.properties.clear_history_on_save:
                self.clear_histories()

            self.log(f"Device saved to {self.properties.db_name}.db", level="info")
        except Exception as error:
            self._log.error(f"Error saving to pickle file: {error}")

    async def points_from_sql(self, db_name):
        """
        Retrieve point list from SQL database
        """
        try:
            points = await self._read_from_sql("SELECT * FROM history;", db_name)
            return list(points.columns.values)[1:]
        except Exception:
            self._log.warning(f"No history retrieved from {db_name}.db:")
            return []

    async def his_from_sql(self, db_name, point):
        """
        Retrive point histories from SQL database
        """
        his = await self._read_from_sql('select * from "history"', db_name)
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[point]

    async def value_from_sql(self, db_name, point):
        """
        Take last known value as the value
        """
        return await self.his_from_sql(db_name, point).last_valid_index()

    def read_point_prop(self, device_name, point):
        """
        Points properties retrieved from pickle
        """
        with open(f"{device_name}.bin", "rb") as file:
            try:
                _point = pickle.load(file)["points"][point]
            except KeyError:
                raise RemovedPointException(f"{point} not found (probably deleted)")
            return _point

    def read_dev_prop(self, device_name):
        """
        Device properties retrieved from pickle
        """
        self.log("Reading prop from DB file", level="debug")
        try:
            with open(f"{device_name}.bin", "rb") as file:
                return pickle.load(file)["device"]
        except (EOFError, FileNotFoundError):
            self._log.error("Error reading device properties")
            raise ValueError



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Read.py - creation of ReadProperty and ReadPropertyMultiple requests

    Used while defining an app:
    Example::

        class BasicScript(WhoisIAm, ReadProperty)

    Class::

        ReadProperty()
            def read()
            def readMultiple()

"""

import asyncio
import re

# --- standard Python modules ---
import typing as t

# from bacpypes3.core import deferred
# from bacpypes.iocb import IOCB, TimeoutError
# from bacpypes3.object import get_object_class, registered_object_types
from bacpypes3.apdu import (
    AbortPDU,
    AbortReason,
    ErrorRejectAbortNack,
    PropertyReference,
    Range,
    ReadAccessSpecification,
    ReadPropertyMultipleRequest,
    ReadRangeACK,
    ReadRangeRequest,
    RejectPDU,
    RejectReason,
)
from bacpypes3.app import Application
from bacpypes3.basetypes import (
    DateTime,
    EngineeringUnits,
    PropertyIdentifier,
    RangeByPosition,
    RangeBySequenceNumber,
    RangeByTime,
)
from bacpypes3.errors import NoResponse, ObjectError
from bacpypes3.object import get_vendor_info

# --- 3rd party modules ---
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import Date, ObjectIdentifier, Tag, Time

from BAC0.core.app.asyncApp import BAC0Application

from ..utils.notes import note_and_log

# --- this application's modules ---
from .IOExceptions import (
    ApplicationNotStarted,
    NoResponseFromController,
    ReadRangeException,
    SegmentationNotSupported,
    UnknownObjectError,
    UnknownPropertyError,
    UnrecognizedService,
)

# ------------------------------------------------------------------------------


ReadValue = t.Union[float, str, t.List]
rpm_request_pattern = r"(?P<request>(?P<Object>[0-9A-Za-z-]+:\d+)[, ]+[(\[ ](?P<Properties>(?P<Property>[0-9A-Za-z-]+(\[\d+\])*[, ]*)+)[)\]]*)"


@note_and_log
class ReadProperty:
    """
    Defines BACnet Read functions: readProperty and readPropertyMultiple.
    Data exchange is made via a Queue object
    A timeout of 10 seconds allows detection of invalid device or communciation errors.
    """

    async def read(
        self,
        args: str,
        arr_index: t.Optional[int] = None,
        vendor_id: int = 0,
        bacoid=None,
        timeout: int = 10,
        show_property_name: bool = False,
    ) -> t.Union[ReadValue, t.Tuple[ReadValue, str], None]:
        """
        Build a ReadProperty request, wait for the answer and return the value

        :param args: String with <addr> <type> <inst> <prop> [ <indx> ]
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.read('2:5 analogInput 1 presentValue')

        Requests the controller at (Network 2, address 5) for the presentValue of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        args_split = args.split()

        (
            device_address,
            object_identifier,
            property_identifier,
            property_array_index,
        ) = self.build_rp_request(
            args_split, arr_index=arr_index, vendor_id=vendor_id, bacoid=bacoid
        )

        self.log_title("Read property", args_split)
        # Do I know you ?
        dic = await self.this_application.app.device_info_cache.get_device_info(
            device_address
        )
        if dic is None:
            _iam = await self.this_application.app.who_is(address=device_address)
            failures = 0
            while _iam == []:
                # retry
                failures += 1
                await asyncio.sleep(1)
                _iam = await self.this_application.app.who_is(address=device_address)
                if failures > 5:
                    self.log(
                        f"Trouble with Iam... Response received from {device_address} = {_iam}",
                        level="error",
                    )
                    raise NoResponseFromController
            await self.this_application.app.device_info_cache.set_device_info(_iam[0])
            dic = await self.this_application.app.device_info_cache.get_device_info(
                device_address
            )
            self.log(f"Device Info Cache : {dic}", level="debug")
        try:
            response = await _app.read_property(
                device_address,
                object_identifier,
                property_identifier,
                property_array_index,
            )

        except ErrorRejectAbortNack as err:
            response = err

            if "unknown-property" in str(err.reason):
                if "description" in args:
                    self._log.warning(
                        "The description property is not implemented in the device. "
                        "Using a default value for internal needs."
                    )
                    return "n/a"
                elif "inactiveText" in args:
                    self._log.warning(
                        "The inactiveText property is not implemented in the device. "
                        "Using a default value of Off for internal needs."
                    )
                    return "False"
                elif "activeText" in args:
                    self._log.warning(
                        "The activeText property is not implemented in the device. "
                        "Using a default value of On for internal needs."
                    )
                    return "True"
                if "units" in args:
                    self._log.warning(
                        "The units property is not implemented in the device. We will consider noUnits"
                        "Using a default value for internal needs. Please note that units is a required property for BACnet objects like analog values. The device you are reading from may be non-compliant."
                    )
                    return EngineeringUnits("noUnits")
                else:

                    raise UnknownPropertyError(f"Unknown property {args}")
            else:
                self.log(f"Error : {err}", level="error")
        except ObjectError:
            raise UnknownObjectError(f"Unknown object {args}")

        # except bufferOverflow
        except NoResponse:
            raise NoResponseFromController

        if not isinstance(response, ErrorRejectAbortNack):
            return response

    def _split_the_read_request(self, args, arr_index):
        """
        When a device doesn't support segmentation, this function
        will split the request according to the length of the
        predicted result which can be known when reading the array_index
        number 0.

        This can be a very long process as some devices count a large
        number of properties without supporting segmentation
        (FieldServers are a good example)
        """
        # parameter arr_index appears to be unused in this function?
        nmbr_obj = self.read(args, arr_index=0)
        return [self.read(args, arr_index=i) for i in range(1, nmbr_obj + 1)]  # type: ignore

    async def readMultiple(
        self,
        args: str,
        request_dict=None,
        vendor_id: int = 0,
        timeout: int = 10,
        show_property_name: bool = False,
        from_regex=False,
    ) -> t.Union[t.Dict, t.List[t.Tuple[t.Any, str]]]:
        """Build a ReadPropertyMultiple request, wait for the answer and return the values

        :param args: String with <addr> ( <type> <inst> ( <prop> [ <indx> ] )... )...
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.readMultiple('2:5 analogInput 1 presentValue units')

        Requests the controller at (Network 2, address 5) for the (presentValue and units) of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        if request_dict is not None:
            address, parameter_list = await self.build_rpm_request_from_dict(
                request_dict, vendor_id
            )
        elif from_regex:
            address, parameter_list = await self.build_rpm_request_from_regex(
                args, vendor_id=vendor_id
            )
            self.log_title(f"Read Multiple for {address} | params : {parameter_list}")

        else:
            args_list = args.split()
            address, parameter_list = await self.build_rpm_request(
                args_list, vendor_id=vendor_id
            )
            self.log_title("Read Multiple", args_list)

        # Force DeviceInfoCache
        dic = await self.this_application.app.device_info_cache.get_device_info(address)
        if dic is None:
            _iam = await self.this_application.app.who_is(address=address)
            await self.this_application.app.device_info_cache.set_device_info(_iam[0])
            dic = await self.this_application.app.device_info_cache.get_device_info(
                address
            )
            self.log(f"Device Info Cache : {dic}", level="debug")

        values = []
        dict_values = {}

        self.log(f"Parameter list : {parameter_list}", level="debug")

        try:
            # build an ReadPropertyMultiple request
            response = await _app.read_property_multiple(address, parameter_list)
            self.log(f"Response : {response}", level="debug")

        except ErrorRejectAbortNack as err:
            # construction error
            response = err
            self._log.exception(f"exception: {err.reason}")
            if "segmentation-not-supported" in str(err.reason):
                raise SegmentationNotSupported
            if "unrecognized-service" in str(err.reason):
                raise UnrecognizedService()
            if "unknown-object" in str(err.reason):
                self.log(f"Unknown object {args}", level="warning")
                raise UnknownObjectError(f"Unknown object {args}")
            if "unknown-property" in str(err.reason):
                values.append("")  # type: ignore[arg-type]
                return values
            if "no-response" in str(err.reason):
                # values.append("")  # type: ignore[arg-type]
                # return values
                # try again
                try:
                    response = await _app.read_property_multiple(
                        address, parameter_list
                    )
                except ErrorRejectAbortNack as err:
                    raise err

        if not isinstance(response, ErrorRejectAbortNack):
            """
            TODO : Need improvement here and look for the property identifier that is coming from the response
            Then we'll be able to support multiple properties for the same object in the read multiple function
            """
            for (
                object_identifier,
                property_identifier,
                property_array_index,
                property_value,
            ) in response:
                self._log.debug(
                    "{!r:<20} {!r:<20} {!r:<30} {!r:<20}".format(
                        property_identifier,
                        property_array_index,
                        property_value,
                        # datatype,
                        "",
                    )
                )
                if str(object_identifier) not in dict_values:
                    dict_values[str(object_identifier)] = []
                if show_property_name:
                    values.append((property_value, property_identifier))
                    dict_values[str(object_identifier)].append(
                        (property_identifier, (property_value, property_identifier))
                    )
                else:
                    values.append(property_value)
                    dict_values[str(object_identifier)].append(
                        (property_identifier, property_value)
                    )

            if request_dict is not None:
                return dict_values
            else:
                return values

        return values

    def build_rp_request(
        self, args: t.List[str], arr_index=None, vendor_id: int = 0, bacoid=None
    ) -> t.Tuple:
        vendor = get_vendor_info(vendor_id)
        try:
            addr, obj_type_str, obj_inst_str, prop_id_str = args[:4]
            object_identifier = ObjectIdentifier((obj_type_str, int(obj_inst_str)))
        except ValueError:
            addr, obj_type_str, prop_id_str = args[:3]
            object_identifier = ObjectIdentifier(obj_type_str)

        device_address = Address(addr)

        # TODO : This part needs work to find proprietary objects
        # obj_type = self.__get_obj_type(obj_type_str, vendor_id)

        if prop_id_str.isdigit():
            prop_id = int(prop_id_str)
        elif "@prop_" in prop_id_str or "@idx" in prop_id_str:
            if "@idx" in prop_id_str:
                prop_id, arr_index = prop_id_str.split("@idx:")
            if "@prop_" in prop_id_str:
                prop_id = int(prop_id_str.split("_")[1])
        else:
            prop_id = prop_id_str  # type: ignore
        prop_id = PropertyIdentifier(prop_id)

        if arr_index is None:
            arr_index = int(args[4]) if len(args) == 5 else arr_index
        params = (device_address, object_identifier, prop_id, arr_index)
        self.log(f"{'REQUEST':<20} {params!r}", level="debug")
        return params

    async def build_rpm_request(
        self, args: t.List[str], vendor_id: int = 0
    ) -> ReadPropertyMultipleRequest:
        """
        Build request from args
        """
        property_array_index = None
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        self.log(args, level="debug")

        vendor_id = vendor_id
        address = Address(args.pop(0))
        # get information about the device from the cache
        device_info = await _app.device_info_cache.get_device_info(address)

        # using the device info, look up the vendor information
        if device_info:
            vendor_info = get_vendor_info(device_info.vendor_identifier)
        else:
            vendor_info = get_vendor_info(vendor_id)

        parameter_list = []
        while args:
            # get the object identifier and using the vendor information, look
            # up the class
            obj_id = args.pop(0)
            if obj_id.isdigit():
                obj_id = int(obj_id)
            elif "@obj_" in obj_id:
                obj_id = int(obj_id.split("@obj_")[1])
            if ":" not in str(obj_id):
                obj_instance = args.pop(0)
            else:
                obj_instance = obj_id.split(":")[1]
                obj_id = obj_id.split(":")[0]
            object_identifier = vendor_info.object_identifier((obj_id, obj_instance))
            object_class = vendor_info.get_object_class(object_identifier[0])
            if not object_class:
                await self.response(f"unrecognized object type: {object_identifier}")
                return

            properties_list = []
            while args:
                # now get the property type from the class
                if "@obj_" in args[0]:
                    break
                elif "@prop_" in args[0] or "@idx" in args[0]:
                    if "@idx" in args[0]:
                        prop_id, arr_index = args.pop(0).split("@idx_")
                    else:
                        prop_id = int(args.pop(0).split("prop_")[1])
                else:
                    prop_id = args.pop(0)
                try:
                    property_identifier = vendor_info.property_identifier(prop_id)
                except ValueError:
                    try:
                        property_identifier = PropertyIdentifier(prop_id)
                    except ValueError:
                        break  # probably another object

                if property_identifier not in (
                    PropertyIdentifier.all,
                    PropertyIdentifier.required,
                    PropertyIdentifier.optional,
                    # "objectName",
                    # "objectType",
                    # "objectIdentifier",
                    # "polarity",
                ):
                    property_type = object_class.get_property_type(property_identifier)
                    if not property_type:
                        await _app.response(
                            f"unrecognized property: {property_identifier}"
                        )
                        return

                # check for a property array index
                if args and args[0].isdigit() and arr_index is None:
                    property_array_index = int(args.pop(0))
                    # save this as a parameter
                    properties_list.append((property_identifier, property_array_index))
                elif property_array_index is not None:
                    properties_list.append((property_identifier, arr_index))
                else:
                    properties_list.append(property_identifier)

                # crude check to see if the next thing is an object identifier
                if args and ((":" in args[0]) or ("," in args[0]) or ("-" in args[0])):
                    break
            parameter_list.append(object_identifier)
            parameter_list.append(properties_list)

        if not parameter_list:
            await _app.response("object identifier expected")
            return
        else:
            return (address, parameter_list)

    async def build_rpm_request_from_regex(self, args, vendor_id=0):
        pattern = re.compile(rpm_request_pattern)
        address = Address(args.split()[0])

        result = re.findall(pattern, args)
        request = []
        for each in result:
            _, object_identifier, properties, _, _ = each
            request.append(ObjectIdentifier(object_identifier))
            request.append([PropertyReference(x) for x in properties.split()])
        self.log(f"RPM Request from Regex : {address} | {request}", level="debug")
        return (address, request)

    async def build_rpm_request_from_dict(self, request_dict, vendor_id=0):
        """
        Read property multiple allow to read a lot of properties with only one request
        The existing RPM function is made using a string that must be created using bacpypes
        console style and is hard to automate.

        This new version will be an attempt to improve that::

            _rpm = {'address': '11:2',
                'objects': {'analogInput:1': ['presentValue', 'description', 'unit', 'objectList@idx:0'],
                            'analogInput:2': ['presentValue', 'description', 'unit', 'objectList@idx:0'],
                },
                vendor_id: 842
                }

        """

        vendor_id = vendor_id
        address = Address(request_dict["address"])
        objects = request_dict["objects"]
        if "vendor_id" in request_dict.keys():
            vendor_id = int(request_dict["vendor_id"])

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        # get information about the device from the cache
        device_info = await _app.device_info_cache.get_device_info(address)

        # using the device info, look up the vendor information
        if device_info:
            vendor_info = get_vendor_info(device_info.vendor_identifier)
        else:
            vendor_info = get_vendor_info(0)

        parameter_list = []
        arr_index = None

        for obj, list_of_properties in objects.items():
            object_identifier = ObjectIdentifier(obj)
            properties_list = []
            for each in list_of_properties:
                if "@obj_" in each:
                    break
                elif "@prop_" in each or "@idx" in each:
                    if "@idx_" in each:
                        prop_id, arr_index = each.split("@idx_")
                    else:
                        prop_id = int(each.split("@prop_")[1])  # there was no idx
                else:
                    prop_id = each
                property_identifier = vendor_info.property_identifier(prop_id)

                if arr_index:
                    properties_list.append(property_identifier, arr_index)
                else:
                    properties_list.append(property_identifier)
            parameter_list.append(object_identifier)
            parameter_list.append(properties_list)

        if not parameter_list:
            await _app.response("object identifier expected")
            return

        return (address, parameter_list)

    def build_rrange_request(
        self, args, range_params=None, arr_index=None, vendor_id=0, bacoid=None
    ):
        addr, obj_type, obj_inst, prop_id = args[:4]

        vendor_id = vendor_id
        bacoid = bacoid

        if obj_type.isdigit():
            obj_type = int(obj_type)

        obj_inst = int(obj_inst)

        if prop_id.isdigit():
            prop_id = int(prop_id)

        # build a request
        request = ReadRangeRequest(
            objectIdentifier=(obj_type, obj_inst), propertyIdentifier=prop_id
        )
        request.pduDestination = Address(addr)
        if range_params is not None:
            range_type, first, date, time, count = range_params
            if range_type == "p":
                rbp = RangeByPosition(referenceIndex=int(first), count=int(count))
                request.range = Range(byPosition=rbp)
            elif range_type == "s":
                rbs = RangeBySequenceNumber(
                    referenceSequenceNumber=int(first), count=int(count)
                )
                request.range = Range(bySequenceNumber=rbs)
            elif range_type == "t":
                rbt = RangeByTime(
                    referenceTime=DateTime(date=Date(date), time=Time(time)),
                    count=int(count),
                )
                request.range = Range(byTime=rbt)
            elif range_type == "x":
                # should be missing required parameter
                request.range = Range()
            else:
                raise ValueError(f"unknown range type: {range_type!r}")

        if len(args) == 5:
            request.propertyArrayIndex = int(args[4])
        self.log(f"{'REQUEST':<20} {request!r}", level="debug")
        return request

    async def readRange(
        self,
        args,
        range_params=None,
        arr_index=None,
        vendor_id=0,
        bacoid=None,
        timeout=10,
    ):
        """
        Build a ReadRangeRequest request, wait for the answer and return the value

        :param args: String with <addr> <type> <inst> <prop> [ <indx> ]
        :param range_params: parameters defining how to query the range, a list of five elements
        :returns: data read from device (list of LogRecords)

        range_params: a list of five elements: (range_type: str, first: int, date: str, time: str, count: int)
            range_type: one of ['p', 's', 't']
                        p - RangeByPosition:
                                uses (first, count)
                        s - RangeBySequenceNumber:
                                uses (first, count)
                        t - RangeByTime: Filter by the given time
                                uses (date, time, count)
            first: int, first element when querying by Position or Sequence Number
            date: str, "YYYY-mm-DD" passed to bacpypes.primitivedata.Date constructor
            time: str, "HH:MM:SS" passed to bacpypes.primitivedata.Time constructor
            count: int, number of elements to return, negative numbers reverse direction of search

        *Example*::

            import BAC0
            from bacpypes.basetypes import Date, Time
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip=myIPAddr)

            log_records = bacnet.readRange('2:5 trendLog 1 logBuffer', range_params=('t', None, '2023-05-12', '12:00:00', 2))
            for log_record in log_records:
              print(Date(log_record.timestamp.date), Time(log_record.timestamp.time), log_record.logDatum.realValue)
            # Date(2023-5-12 fri) Time(12:10:00.00) 130.331
            # Date(2023-5-12 fri) Time(12:20:00.00) 134.123

            log_records = bacnet.readRange('2:5 trendLog 1 logBuffer', range_params=('t', None, '2023-05-12', '12:00:00', -2))
            for log_record in log_records:
              print(Date(log_record.timestamp.date), Time(log_record.timestamp.time), log_record.logDatum.realValue)
            # Date(2023-5-12 fri) Time(11:40:00.00) 123.4
            # Date(2023-5-12 fri) Time(11:50:00.00) 125.1213
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        args_split = args.split()

        self.log_title("Read range ", args_split)

        # get information about the device from the cache
        device_info = await _app.device_info_cache.get_device_info(
            Address(args_split[0])
        )

        # using the device info, look up the vendor information
        if device_info:
            vendor_info = get_vendor_info(device_info.vendor_identifier)
        else:
            vendor_info = get_vendor_info(0)

        try:
            # build ReadProperty request
            request = self.build_rrange_request(
                args_split,
                range_params=range_params,
                arr_index=arr_index,
                vendor_id=vendor_id,
                bacoid=bacoid,
            )

            self.log(f"{'request':<20} {request!r}", level="debug")

        except ReadRangeException as error:
            # construction error
            self._log.exception(f"exception: {error!r}")

        response = await _app.request(request)

        if isinstance(response, ErrorRejectAbortNack):
            return response
        if not isinstance(response, ReadRangeACK):
            return None

        object_class = vendor_info.get_object_class(response.objectIdentifier[0])
        datatype = object_class.get_property_type(response.propertyIdentifier)

        value = response.itemData.cast_out(datatype)

        return value

    async def read_priority_array(self, addr, obj, obj_instance) -> t.List:
        pa = await self.read(f"{addr} {obj} {obj_instance} priorityArray")
        res = [pa]
        for each in range(1, 17):
            _pa = pa[each]  # type: ignore[index]
            for k, v in _pa.__dict__.items():
                if v is not None:
                    res.append(v)
        return res


def find_reason(apdu):
    try:
        if apdu is TimeoutError:
            return "Timeout"
        elif apdu.pduType == RejectPDU.pduType:
            reasons = RejectReason.enumerations
        elif apdu.pduType == AbortPDU.pduType:
            reasons = AbortReason.enumerations
        else:
            _code = None
            try:
                _code = apdu.errorCode
            except AttributeError:
                try:
                    _code = apdu.errorType.errorCode
                except AttributeError:
                    raise ValueError("Cannot find reason...")

            if _code:
                return f"{_code}"
        code = apdu.apduAbortRejectReason
        try:
            return [k for k, v in reasons.items() if v == code][0]
        except IndexError:
            return code
    except KeyError as err:
        return f"KeyError: {type(apdu)} has no key {err.args[0]!r}"


def cast_datatype_from_tag(propertyValue, obj_id, prop_id):
    try:
        tag_list = propertyValue.tagList.tagList
        if tag_list[0].tagClass == 0:
            tag = tag_list[0].tagNumber

            datatype = Tag._app_tag_class[tag]
        else:
            from bacpypes3.constructeddata import ArrayOf

            subtype_tag = propertyValue.tagList.tagList[0].tagList[0].tagNumber
            datatype = ArrayOf(Tag._app_tag_class[subtype_tag])
        value = {f"{obj_id}_{prop_id}": propertyValue.cast_out(datatype)}
    except Exception:
        value = {f"{obj_id}_{prop_id}": propertyValue}
    return value


def build_read_access_spec(obj_type, obj_instance, property_reference_list):
    return ReadAccessSpecification(
        objectIdentifier=(obj_type, obj_instance),
        listOfPropertyReferences=property_reference_list,
    )


def build_property_reference_list(obj_type, list_of_properties):
    property_reference_list = []
    for prop in list_of_properties:
        idx = None
        if "@idx:" in prop:
            prop, idx = prop.split("@idx:")
        prop_id = validate_property_id(obj_type, prop)
        prop_reference = PropertyReference(propertyIdentifier=prop_id)
        if idx:
            prop_reference.propertyArrayIndex = int(idx)
        property_reference_list.append(prop_reference)
    return property_reference_list


def validate_property_id(obj_type, prop_id):
    if prop_id in PropertyIdentifier.enumerations:
        if prop_id in (
            "all",
            "required",
            "optional",
            "objectName",
            "objectType",
            "objectIdentifier",
            "polarity",
        ):
            return prop_id
        # elif validate_datatype(obj_type, prop_id) is not None:
        #    return prop_id
        else:
            raise ValueError(
                f"invalid property for object type : {obj_type} | {prop_id}"
            )
    elif "@prop_" in prop_id:
        return int(prop_id.split("_")[1])
    else:
        raise ValueError(f"{prop_id} is an invalid property for {obj_type}")


# def validate_datatype(obj_type, prop_id, vendor_id=842):
#    return get_datatype(obj_type, prop_id, vendor_id=vendor_id) if not None else False



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Points.py - Definition of points so operations on Read results are more convenient.
"""

import asyncio
import typing as t
from collections import namedtuple

# --- standard Python modules ---
from datetime import datetime, timedelta

from bacpypes3.basetypes import BinaryPV, PropertyIdentifier
from bacpypes3.pdu import Address

# --- 3rd party modules ---
from bacpypes3.primitivedata import Boolean, CharacterString, ObjectIdentifier

from ...scripts.Base import Base
from ...tasks.Match import Match, Match_Value

# --- this application's modules ---
from ...tasks.Poll import SimplePoll as Poll
from ..io.IOExceptions import (
    NoResponseFromController,
    RemovedPointException,
    UnknownPropertyError,
    WritePropertyException,
)
from ..utils.lookfordependency import pandas_if_available
from ..utils.notes import note_and_log

_PANDAS, pd, sql, Timestamp = pandas_if_available()
# ------------------------------------------------------------------------------


class PointProperties(object):
    """
    A container for point properties.
    """

    def __init__(self):
        self.device = None
        self.name = None
        self.type = ""
        self.address = -1
        self.description = None
        self.units_state = None
        self.simulated: t.Tuple[bool, t.Optional[int]] = (False, None)
        self.overridden: t.Tuple[bool, t.Optional[int]] = (False, None)
        self.priority_array = None
        self.history_size = None
        self.bacnet_properties = {}
        self.status_flags = None

    def __repr__(self):
        return f"{self.asdict}"

    @property
    def asdict(self):
        return self.__dict__


# ------------------------------------------------------------------------------


@note_and_log
class Point:
    """
    Represents a device BACnet point.  Used to NumericPoint, BooleanPoint and EnumPoints.

    Each point implements a history feature. Each time the point is read, its value (with timestamp)
    is added to a history table. Histories capture the changes to point values over time.
    """

    _cache_delta = timedelta(seconds=5)

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
        tags=[],
    ):
        self._history = namedtuple("_history", ["timestamp", "value"])
        self.properties = PointProperties()

        self._polling_task = namedtuple("_polling_task", ["task", "running"])
        self._polling_task.task = None
        self._polling_task.running = False

        self._match_task = namedtuple("_match_task", ["task", "running"])
        self._match_task.task = None
        self._match_task.running = False

        self._history.timestamp = []
        self._history.value = []
        # self._history.value = [presentValue]
        # self._history.timestamp.append(datetime.now().astimezone())

        self.properties.history_size = history_size

        self.properties.device = device
        self.properties.name = str(pointName)
        self.properties.type = pointType
        self.properties.address = pointAddress

        self.properties.description = str(description)
        self.properties.units_state = units_state
        self.properties.simulated = (False, 0)
        self.properties.overridden = (False, 0)

        self.cov_registered = False

        self.tags = tags

        self._cache: t.Dict[str, t.Tuple[t.Optional[datetime], t.Any]] = {
            "_previous_read": (None, None)
        }

    @property
    async def value(self):
        """
        Retrieve value of the point
        """
        if (
            self._cache["_previous_read"][0]
            and datetime.now().astimezone() - self._cache["_previous_read"][0]
            < Point._cache_delta
        ):
            return self._cache["_previous_read"][1]

        try:
            res = await self.properties.device.properties.network.read(
                "{} {} {} presentValue".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                ),
                vendor_id=self.properties.device.properties.vendor_id,
            )
            # self._trend(res)
        except Exception:
            raise
        self._cache["_previous_read"] = (datetime.now().astimezone(), res)
        return res

    async def read_priority_array(self):
        """
        Retrieve priority array of the point
        """
        if self.properties.priority_array is not False:
            try:
                res = await self.properties.device.properties.network.read(
                    "{} {} {} priorityArray".format(
                        self.properties.device.properties.address,
                        self.properties.type,
                        str(self.properties.address),
                    ),
                    vendor_id=self.properties.device.properties.vendor_id,
                )
                self.properties.priority_array = []
                for i, each in enumerate(res):
                    _t = each.__dict__["_choice"]
                    val = each.__dict__[_t]
                    self.properties.priority_array.append(
                        {
                            "priority": i + 1,
                            "priorityValue": each,
                            "value": val,
                            "choice": _t,
                        }
                    )
            except (ValueError, UnknownPropertyError):
                self.properties.priority_array = False
            except Exception as e:
                raise Exception(f"Problem reading : {self.properties.name} | {e}")

    async def read_property(self, prop):
        try:
            return await self.properties.device.properties.network.read(
                "{} {} {} {}".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                    prop,
                ),
                vendor_id=self.properties.device.properties.vendor_id,
            )
        except Exception as e:
            raise Exception(f"Problem reading : {self.properties.name} | {e}")

    async def update_bacnet_properties(self):
        """
        Retrieve bacnet properties for this point
        To retrieve something general, forcing vendor id 0
        """
        try:
            res = await self.properties.device.properties.network.readMultiple(
                "{} {} {} all".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                ),
                vendor_id=self.properties.device.properties.vendor_id,
                show_property_name=True,
            )
            for each in res:
                if not each:
                    continue
                v, prop = each
                self.properties.bacnet_properties[prop] = v

        except Exception as e:
            raise Exception(f"Problem reading : {self.properties.name} | {e}")

    @property
    async def bacnet_properties(self):
        if not self.properties.bacnet_properties:
            await self.update_bacnet_properties()
        return self.properties.bacnet_properties

    @property
    def is_overridden(self):
        self.read_priority_array()
        if self.properties.priority_array is False:
            return False
        if self.priority(8) or self.priority(1):
            self.properties.overridden = (True, self.value)
            return True
        else:
            return False

    async def priority(self, priority=None):
        if self.properties.priority_array is False:
            return None

        await self.read_priority_array()
        if priority is None:
            return self.properties.priority_array
        if priority < 1 or priority > 16:
            raise IndexError("Please provide priority to read (1-16)")

        else:
            val = self.properties.priority_array[priority - 1]["value"]._value
            if isinstance(val, tuple) and len(val) == 0:
                return None
            return val

    def _trend(self, res: t.Optional[t.Union[float, int, str]]) -> None:
        now = datetime.now().astimezone()
        self._history.timestamp.append(now)
        self._history.value.append(res)
        if self.properties.device.properties.network.database:
            self.properties.device.properties.network.database.prepare_point([self])

        if self.properties.history_size is None:
            return
        else:
            if self.properties.history_size < 1:
                self.properties.history_size = 1
            if len(self._history.timestamp) >= self.properties.history_size:
                try:
                    self._history.timestamp = self._history.timestamp[
                        -self.properties.history_size :  # noqa E203
                    ]
                    self._history.value = self._history.value[
                        -self.properties.history_size :  # noqa E203
                    ]
                    assert len(self._history.timestamp) == len(self._history.value)

                except Exception:
                    self._log.exception("Can't append to history")

    @property
    def units(self):
        """
        Should return units
        """
        raise Exception("Must be overridden")

    @property
    def lastValue(self):
        """
        returns: last value read
        """
        if _PANDAS:
            last_val = self.history.dropna()
            last_val_clean = None if len(last_val) == 0 else last_val.iloc[-1]
            return last_val_clean
        else:
            return self._history.value[-1]

    @property
    def lastTimestamp(self):
        """
        returns: last timestamp read
        """
        if _PANDAS:
            last_val = self.history.dropna()
            last_val_clean = None if len(last_val) == 0 else last_val.index[-1]
            return last_val_clean
        else:
            return self._history.timestamp[-1]

    @property
    def history(self) -> t.Dict[datetime, t.Union[int, float, str]]:
        """
        returns : (pd.Series) containing timestamp and value of all readings
        """
        if not _PANDAS:
            return dict(zip(self._history.timestamp, self._history.value))
        idx = self._history.timestamp.copy()
        his_table = pd.Series(index=idx, data=self._history.value[: len(idx)])
        del idx
        his_table.name = ("{}/{}").format(
            self.properties.device.properties.name, self.properties.name
        )
        his_table.units = self.properties.units_state
        if self.properties.name in self.properties.device.binary_states:
            his_table.states = "binary"
        elif self.properties.name in self.properties.device.multi_states:
            his_table.states = "multistates"
        else:
            his_table.states = "analog"
        his_table.description = self.properties.description

        his_table.datatype = self.properties.type
        return his_table

    def clear_history(self):
        self._history.timestamp = []
        self._history.value = []

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        if remove:
            self.properties.device.properties.network.remove_trend(self)
        else:
            self.properties.device.properties.network.add_trend(self)

    @property
    def status(self):
        return self.properties.status_flags

    async def __getitem__(self, key):
        """
        Way to get points... presentValue, status, flags, etc...

        :param key: state
        :returns: list of enum states
        """
        if str(key).lower() in ["unit", "units", "state", "states"]:
            key = "units_state"
        try:
            return getattr(self.properties, key)
        except AttributeError:
            try:
                if "@prop_" in key:
                    key = key.split("prop_")[1]
                return await self.read_property(key)
            except Exception:
                raise ValueError(f"Cannot find property named {key}")

    async def write(
        self, value: t.Any, *, prop: str = "presentValue", priority: str = "16"
    ):
        """
        Write to present value of a point

        :param value: (float) numeric value
        :param prop: (str) property to write. Default = presentValue
        :param priority: (int) priority to which write.

        """
        if prop == "description":
            await self.update_description(value)
        else:
            if priority != "":
                if (
                    isinstance(float(priority), float)
                    and float(priority) >= 1
                    and float(priority) <= 16
                ):
                    priority = f"{priority}"
                else:
                    raise ValueError("Priority must be a number between 1 and 16")
            req = f"{self.properties.device.properties.address} {self.properties.type} {self.properties.address} {prop} {value} - {priority}"
            # self.log(req, level='info')
            try:
                response = await self.properties.device.properties.network._write(
                    req,
                    vendor_id=self.properties.device.properties.vendor_id,
                )
                # print(response)
                self.log(f"Write response : {response}", level="debug")
            except NoResponseFromController:
                raise

            # Read after the write so history gets updated.
            await self.value

    async def default(self, value):
        await self.write(value, prop="relinquishDefault")

    async def sim(self, value, *, force=False):
        """
        Simulate a value.  Sets the Out_Of_Service property- to disconnect the point from the
        controller's control.  Then writes to the Present_Value.
        The point name is added to the list of simulated points (self.simPoints)

        :param value: (float) value to simulate
        """
        if (
            not self.properties.simulated[0]
            or self.properties.simulated[1] != value
            or force is not False
        ):
            await self.properties.device.properties.network.sim(
                f"{self.properties.device.properties.address} {self.properties.type} {self.properties.address} presentValue {value}",
                vendor_id=self.properties.device.properties.vendor_id,
            )
            self.properties.simulated = (True, value)

    async def out_of_service(self):
        """
        Sets the Out_Of_Service property [to True].
        """
        await self.properties.device.properties.network.out_of_service(
            f"{self.properties.device.properties.address} {self.properties.type} {self.properties.address} outOfService True"
        )
        self.properties.simulated = (True, None)

    async def is_out_of_service(self):
        """
        Check if the Out_Of_Service property is true.
        """
        res = await self.properties.device.properties.network.is_out_of_service(
            f"{self.properties.device.properties.address} {self.properties.type} {self.properties.address} outOfService"
        )
        self.properties.simulated = (True, None)
        return res

    async def release(self):
        """
        Clears the Out_Of_Service property [to False] - so the controller regains control of the point.
        """
        await self.properties.device.properties.network.release(
            f"{self.properties.device.properties.address} {self.properties.type} {self.properties.address} outOfService inactive"
        )
        self.properties.simulated = (False, None)

    async def ovr(self, value):
        await self.write(value, priority=8)
        self.properties.overridden = (True, value)

    async def auto(self):
        await self.write("null", priority=8)
        self.properties.overridden = (False, 0)

    async def release_ovr(self):
        await self.write("null", priority=1)
        await self.write("null", priority=8)
        self.properties.overridden = (False, None)

    async def _setitem(self, value):
        """
        Called by _set, will trigger right function depending on
        point type to write to the value and make tests.
        This is default behaviour of the point  :
        AnalogValue are written to
        AnalogOutput are overridden
        """
        self.log(f"Setting to {value}", level="debug")
        if "characterstring" in self.properties.type.lower():
            await self.write(value)

        elif "value" in self.properties.type.lower():
            if str(value).lower() == "auto":
                raise ValueError(
                    "Value was not simulated or overridden, cannot release to auto"
                )
            # analog value must be written to
            await self.write(value)

        elif "output" in self.properties.type.lower():
            # analog output must be overridden
            if str(value).lower() == "auto":
                await self.auto()
            else:
                await self.ovr(value)
        else:
            # input are left... must be simulated
            if str(value).lower() == "auto":
                await self.release()
            else:
                self.log(f"Simulating to {value}", level="debug")
                await self.sim(value)

    async def _set(self, value):
        """
        Allows the syntax:
            device['point'] = value
        """
        raise NotImplementedError("Must be overridden")

    def poll(self, command="start", *, delay: int = 10) -> None:
        asyncio.create_task(self._poll(command=command, delay=delay))

    async def _poll(self, command="start", *, delay: int = 10) -> None:
        """
        Poll a point every x seconds (delay=x sec)
        Stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0
        """
        if (
            str(command).lower() == "stop"
            or command is False
            or command == 0
            or delay == 0
        ):
            if isinstance(self._polling_task.task, Poll):
                self._polling_task.task.stop()
                self._polling_task.task = None
                self._polling_task.running = False

        elif self._polling_task.task is None:
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        elif self._polling_task.running:
            self._polling_task.task.stop()
            self._polling_task.running = False
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        else:
            raise RuntimeError("Stop polling before redefining it")

    def match(self, point, *, delay=5):
        asyncio.create_task(self._match(point=point, delay=delay))

    async def _match(self, point, *, delay=5):
        """
        This allow functions like :
            device['status'].match('command')

        A fan status for example will follow the command...
        """
        if self._match_task.task is None:
            self._match_task.task = Match(command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            await self._match_task.task.stop()
            self._match_task.running = False
            await asyncio.sleep(1)

            self._match_task.task = Match(command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            await self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def match_value(self, value, *, delay=5, use_last_value=False):
        asyncio.create_task(
            self._match_value(value=value, delay=delay, use_last_value=use_last_value)
        )

    async def _match_value(self, value, *, delay=5, use_last_value=False):
        """
        This allow functions like :
            device['point'].match('value')

        A sensor will follow a calculation...
        """
        if self._match_task.task is None:
            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay, use_last_value=use_last_value
            )
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            await self._match_task.task.stop()
            self._match_task.running = False
            await asyncio.sleep(1)

            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay, use_last_value=use_last_value
            )
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            await self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def __len__(self):
        """
        Length of a point = # of history records
        """
        return len(self.history)

    async def subscribe_cov(
        self, confirmed: bool = False, lifetime: int = 900, callback=None
    ):
        """
        Subscribes to the Change of Value (COV) service for this point.

        The COV service allows the device to notify the application of changes to the value of a property.
        This method sets up the subscription and starts an asynchronous task to listen for these notifications.

        Args:
            confirmed (bool, optional): If True, the device will wait for a confirmation from the application
                after sending a COV notification. Defaults to True.
            lifetime (int, optional): The lifetime of the subscription in seconds. If None, the subscription
                will last indefinitely. Defaults to None.
            callback (function, optional): A function to be called when a COV notification is received.
                The function should accept two arguments: the sender (the device) and the args (the notification).

        Raises:
            RuntimeError: If the task is already running, a RuntimeError will be raised.

        Returns:
            None
        """
        self.cov_task = COVPointSubscription(
            point=self, confirmed=confirmed, lifetime=lifetime, callback=callback
        )
        Base._running_cov_tasks[self.cov_task.process_identifier] = self.cov_task
        self.cov_task.task = asyncio.create_task(self.cov_task.run())

    async def cancel_cov(self):
        self.log(f"Canceling COV subscription for {self.properties.name}", level="info")
        process_identifer = self.cov_task.process_identifier

        if process_identifer not in Base._running_cov_tasks:
            await self.response("COV subscription not found")
            return
        cov_subscription = Base._running_cov_tasks.pop(process_identifer)
        cov_subscription.stop()
        # await cov_subscription.task

    def update_description(self, value):
        asyncio.create_task(self._update_description(value=value))

    async def _update_description(self, value):
        """
        This will write to the BACnet point and modify the description
        of the object
        """
        await self.properties.device.properties.network.send_text_write_request(
            addr=self.properties.device.properties.address,
            obj_type=self.properties.type,
            obj_inst=int(self.properties.address),
            value=value,
            prop_id="description",
        )
        self.properties.description = str(self.read_property("description"))

    def tag(self, tag_id, tag_value, lst=None):
        """
        Add tag to point. Those tags can be used to make queries,
        add information, etc.
        They will be included if InfluxDB is used.
        """
        if lst is None:
            self.tag.append((tag_id, tag_value))
        else:
            for each in lst:
                tag_id, tag_value = each
                self.tag.append((tag_id, tag_value))

    async def _update_value(self):
        await asyncio.wait_for(self.value, timeout=1.0)

    def _update_value_if_required(self):
        value_too_old = (
            self._history.timestamp[-1]
            > datetime.now().astimezone() - Point._cache_delta
        )
        if value_too_old:
            try:
                loop = asyncio.get_running_loop()
                loop.create_task(self._update_value())
            except Exception as e:
                self.log(f"Error updating value : {e}", level="error")
                return self.lastValue
        if datetime.now().astimezone() - self._history.timestamp[-1] > timedelta(
            seconds=60
        ):
            self.log(
                f"Last known value {self._history.value[-1]} with timestamp of {self._history.timestamp[-1]}, older than 10sec {datetime.now().astimezone()}. Consider using dev['point'].lastValue if you trust polling of device of manage a up to date read in asynchronous side of your app for better precision",
                level="warning",
            )
        return self.lastValue


# ------------------------------------------------------------------------------


class NumericPoint(Point):
    """
    Representation of a Numeric value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):
        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )

    #    @property
    #    def val(self):
    #        res = asyncio.run_coroutine_threadsafe(self.value)
    #        return res.result()

    @property
    async def value(self):
        res = await super().value
        self._trend(res)
        return res

    @property
    def units(self):
        return self.properties.units_state

    async def _set(self, value):
        if str(value).lower() == "auto":
            await self._setitem(value)
        else:
            try:
                if isinstance(value, Point):
                    value = value.lastValue
                val = float(value)
                if isinstance(val, float):
                    await self._setitem(value)
            except Exception as error:
                raise WritePropertyException(f"Problem writing to device : {error}")

    def __repr__(self):
        val = self.lastValue if self.lastValue is not None else float("nan")
        return f"{self.properties.device.properties.name}/{self.properties.name} : {val:.2f} {self.properties.units_state}"

    def __add__(self, other):
        self._update_value_if_required()
        return self.lastValue + other

    __radd__ = __add__

    def __sub__(self, other):
        return self._update_value_if_required() - other

    def __rsub__(self, other):
        return other - self._update_value_if_required()

    def __mul__(self, other):
        return self._update_value_if_required() * other

    __rmul__ = __mul__

    def __truediv__(self, other):
        return self._update_value_if_required() / other

    def __rtruediv__(self, other):
        return other / self._update_value_if_required()

    def __lt__(self, other):
        return self._update_value_if_required() < other

    def __le__(self, other):
        return self._update_value_if_required() <= other

    def __eq__(self, other):
        return self._update_value_if_required() == other

    def __gt__(self, other):
        return self._update_value_if_required() > other

    def __ge__(self, other):
        return self._update_value_if_required() >= other


# ------------------------------------------------------------------------------


class BooleanPoint(Point):
    """
    Representation of a Boolean value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):
        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )
        self.properties.units_state = tuple(str(x) for x in units_state)

    def _trend(self, res):
        if res is not None:
            res = "1: active" if res == BinaryPV.active else "0: inactive"
        super()._trend(res)

    @property
    async def value(self):
        """
        Read the value from BACnet network
        """
        res = await super().value
        self._trend(res)

        if res == BinaryPV.inactive:
            self._key = 0
            self._boolKey = False
        else:
            self._key = 1
            self._boolKey = True
        return res

    @property
    def _boolValue(self):
        """
        returns : (boolean) Value
        """
        if ":" in self.lastValue:
            _val = int(self.lastValue.split(":")[0])
        else:
            _val = self.lastValue
        if _val in [1, "active"]:
            self._key = 1
            self._boolKey = True
        else:
            self._key = 0
            self._boolKey = False
        return self._boolKey

    @property
    async def boolValue(self):
        """
        returns : (boolean) Value
        """
        await self.value  # Force a read
        return self._boolValue

    @property
    def units(self):
        """
        Boolean points don't have units
        """
        return None

    async def _set(self, value):
        try:
            if value is True:
                await self._setitem("active")
            elif value is False:
                await self._setitem("inactive")
            elif str(value) in ["inactive", "active"] or str(value).lower() == "auto":
                await self._setitem(value)
            else:
                raise ValueError(
                    'Value must be boolean True, False or "active"/"inactive"'
                )
        except (Exception, ValueError) as error:
            raise WritePropertyException(f"Problem writing to device : {error}")

    def __repr__(self):
        val = self._boolValue if self._boolValue is not None else "NaN"
        return (
            f"{self.properties.device.properties.name}/{self.properties.name} : {val}"
        )

    def __or__(self, other):
        self._update_value_if_required()
        return self._boolValue | other

    def __and__(self, other):
        self._update_value_if_required()
        return self._boolValue & other

    def __xor__(self, other):
        self._update_value_if_required()
        return self._boolValue ^ other

    def __eq__(self, other):
        self._update_value_if_required()
        if isinstance(other, str):
            if ":" in other:
                return self.lastValue == other
            elif other in ("inactive", "active"):
                return self.lastValue.split(":")[1] == other
            else:
                return False
        elif isinstance(other, bool):
            return self._boolValue == other
        elif isinstance(other, int):
            return int(self.lastValue.split(":")[0]) == other
        else:
            return False


# ------------------------------------------------------------------------------


class EnumPoint(Point):
    """
    Representation of an Enumerated (multiState) value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):
        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )
        self.properties.units_state = (
            [str(x) for x in units_state] if units_state else []
        )

    def _trend(self, res):
        if res is not None:
            res = f"{res}: {self.get_state(res)}"
        super()._trend(res)

    @property
    async def value(self):
        res = await super().value
        # self.log("Value : {}".format(res), level='info')
        # self.log("EnumValue : {}".format(self.get_state(res)), level='info')
        self._trend(res)
        return res

    def get_state(self, v):
        try:
            # errors caught below
            return self.properties.units_state[v - 1]  # type: ignore[index]
        except (TypeError, IndexError):
            return "n/a"

    @property
    def _enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            if ":" in self.lastValue:
                _val = int(self.lastValue.split(":")[0])
                return self.get_state(_val)
        except TypeError:
            # probably first occurence of history... retry
            return self.get_state(self.lastValue)
        except IndexError:
            value = "unknown"
        except ValueError:
            value = "NaN"
        return value

    @property
    async def enumValue(self):
        """
        returns: (str) Enum state value
        """
        await self.value
        return self._enumValue

    @property
    def units(self):
        """
        Enums have 'state text' instead of units.
        """
        return None

    async def _set(self, value):
        try:
            if isinstance(value, int):
                await self._setitem(value)
            elif str(value) in self.properties.units_state:  # type: ignore[operator]
                await self._setitem(self.properties.units_state.index(value) + 1)
            elif str(value).lower() == "auto":
                await self._setitem("auto")
            else:
                raise ValueError(
                    "Value must be integer or correct enum state : {}".format(
                        self.properties.units_state
                    )
                )
        except (Exception, ValueError) as error:
            raise WritePropertyException(f"Problem writing to device : {error}")

    def __repr__(self):
        val = self._enumValue if self._enumValue is not None else "NaN"
        return (
            f"{self.properties.device.properties.name}/{self.properties.name} : {val}"
        )

    def __eq__(self, other):
        self._update_value_if_required()
        if isinstance(other, str):
            if ":" in other:
                return self.lastValue == other
            else:
                return self._enumValue == other
        elif isinstance(other, int):
            return int(self.lastValue.split(":")[0]) == other
        else:
            return self.lastValue == other


class StringPoint(Point):
    """
    Representation of CharacterString value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        units_state=None,
        presentValue=None,
        history_size=None,
    ):
        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            history_size=history_size,
        )

    @property
    def units(self):
        """
        Characterstring value do not have units or state text
        """
        return None

    def _trend(self, res):
        super()._trend(res)

    @property
    async def value(self):
        res = await super().value
        self._trend(res)
        return res

    async def _set(self, value):
        try:
            if isinstance(value, str):
                await self._setitem(value)
            elif isinstance(value, CharacterString):
                await self._setitem(value.value)
            else:
                raise ValueError("Value must be string or CharacterString")
        except (Exception, ValueError) as error:
            raise WritePropertyException(f"Problem writing to device : {error}")

    def __repr__(self):
        val = self.lastValue if self.lastValue is not None else "NaN"
        return (
            f"{self.properties.device.properties.name}/{self.properties.name} : {val}"
        )

    def __eq__(self, other):
        self._update_value_if_required()
        return self.lastValue == other.value


class DateTimePoint(Point):
    """
    Representation of DatetimeValue value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        units_state=None,
        presentValue=None,
        history_size=None,
    ):
        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            history_size=history_size,
        )

    @property
    def units(self):
        """
        Characterstring value do not have units or state text
        """
        return None

    def _trend(self, res):
        # super()._trend(res)
        return

    @property
    async def value(self):
        res = await super().value
        # self._trend(res)
        year, month, day, dayofweek = res.date
        hour, minutes, seconds, ms = res.time
        res = datetime(year + 1900, month, day, hour, minutes, seconds)
        return res

    def _set(self, value):
        # try:
        #    if isinstance(value, str):
        #        self._setitem(value)
        #    elif isinstance(value, CharacterString):
        #        self._setitem(value.value)
        #    else:
        #        raise ValueError("Value must be string or CharacterString")
        # except (Exception, ValueError) as error:
        #    raise WritePropertyException("Problem writing to device : {}".format(error))`
        raise NotImplementedError("Writing to Datetime is not supported yet")

    def __repr__(self):
        return f"{self.properties.device.properties.name}/{self.properties.name} : {self.lastValue}"

    def __eq__(self, other):
        self._update_value_if_required()
        return self.lastValue == other.value

    def __ge__(self, other):
        self._update_value_if_required()
        return self.lastValue >= other.value

    def __le__(self, other):
        self._update_value_if_required()
        return self.lastValue <= other.value

    def __gt__(self, other):
        self._update_value_if_required()
        return self.lastValue > other.value

    def __lt__(self, other):
        self._update_value_if_required()
        return self.lastValue < other.value


# ------------------------------------------------------------------------------


class OfflinePoint(Point):
    """
    When offline (DB state), points needs to behave in a particular way
    (we can't read on bacnet...)
    """

    def __init__(self, device, name):
        self.properties = PointProperties()
        self.properties.device = device
        dev_name = self.properties.device.properties.db_name
        try:
            props = self.properties.device.read_point_prop(dev_name, name)
        except RemovedPointException:
            raise

        self.properties.name = props["name"]
        self.properties.type = props["type"]
        self.properties.address = props["address"]

        self.properties.description = props["description"]
        self.properties.units_state = props["units_state"]
        self.properties.simulated = (True, None)
        self.properties.overridden = (False, None)

        if "analog" in self.properties.type:
            self.new_state(NumericPointOffline)
        elif "multi" in self.properties.type:
            self.new_state(EnumPointOffline)
        elif "binary" in self.properties.type:
            self.new_state(BooleanPointOffline)
        elif "string" in self.properties.type:
            self.new_state(StringPointOffline)
        else:
            raise TypeError("Unknown point type")

    def new_state(self, newstate):
        self.__class__ = newstate


class NumericPointOffline(NumericPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "history"',
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = 65535
        return value

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    @property
    def units(self):
        return self.properties.units_state

    def _set(self, value):
        raise OfflineException("Must be online to write")

    def __repr__(self):
        return "{}/{} : {:.2f} {}".format(
            self.properties.device.properties.name,
            self.properties.name,
            self.value,
            self.properties.units_state,
        )


class BooleanPointOffline(BooleanPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "history"',
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class EnumPointOffline(EnumPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "history"',
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        except ValueError:
            value = "NaN"
        return value

    @property
    def enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            value = self.properties.units_state[int(self.lastValue) - 1]  # type: ignore[index]
        except IndexError:
            value = "unknown"
        except ValueError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class StringPointOffline(EnumPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "history"',
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        except ValueError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class COVPointSubscription:
    """
    COVPointSubscription is a class that handles Change of Value (COV) subscriptions for BACnet points.
    It allows subscribing to COV notifications for a specific point, and handles the asynchronous
    communication with the BACnet device to receive these notifications.

    Attributes:
        address (Address): The BACnet address of the device.
        cov_fini (asyncio.Event): An event to signal the end of the COV subscription.
        task (asyncio.Task): The asyncio task that runs the COV subscription.
        obj_identifier (ObjectIdentifier): The BACnet object identifier for the point.
        _app (BACnetApplication): The BACnet application instance.
        process_identifier (int): The process identifier for the COV subscription.
        point (Point): The `BAC0.point` for which the COV subscription is created.
        lifetime (int): The lifetime of the COV subscription in seconds.
        confirmed (bool): Whether the COV notifications should be confirmed.
        callback (Optional[Union[Callable[[str, Any], None], Awaitable[None]]]): The callback function to be called when a COV notification is received.

    Methods:
        __init__(self, point: Point, lifetime: int = 900, confirmed: bool = False, callback: Optional[Union[Callable[[str, Any], None], Awaitable[None]]] = None):
            Initializes the COVPointSubscription instance.

        run(self):
            Asynchronously runs the COV subscription, listening for COV notifications and calling the callback function if provided.

        stop(self):
            Stops the COV subscription by setting the cov_fini event.
    """

    def __init__(
        self,
        point: Point = None,
        lifetime: int = 900,
        confirmed: bool = False,
        callback: t.Optional[
            t.Union[t.Callable[[str, t.Any], None], t.Awaitable[None]]
        ] = None,
    ):
        self.address = Address(point.properties.device.properties.address)
        self.cov_fini = asyncio.Event()
        self.task = None
        self.obj_identifier = ObjectIdentifier(
            (point.properties.type, int(point.properties.address))
        )
        self._app = point.properties.device.properties.network.this_application.app
        self.process_identifier = Base._last_cov_identifier + 1
        Base._last_cov_identifier = self.process_identifier

        self.point = point
        self.lifetime = lifetime
        self.confirmed = confirmed
        self.callback = callback

    async def run(self):
        self.point.cov_registered = True
        self.point.log(
            f"Subscribing to COV for {self.point.properties.name}", level="debug"
        )

        try:
            async with self._app.change_of_value(
                self.address,
                self.obj_identifier,
                self.process_identifier,
                self.confirmed,
                self.lifetime,
            ) as scm:
                cov_fini_task_monitor = asyncio.create_task(self.cov_fini.wait())
                while not self.cov_fini.is_set():
                    incoming: asyncio.Future = asyncio.ensure_future(scm.get_value())
                    done, pending = await asyncio.wait(
                        [
                            incoming,
                        ],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    for task in pending:
                        task.cancel()

                    if incoming in done:
                        property_identifier, property_value = incoming.result()
                        self.point.log(
                            f"COV notification received for {self.point.properties.name} | {property_identifier} -> {type(property_identifier)} with value {property_value} | {property_value} -> {type(property_value)}",
                            level="info",
                        )
                        if property_identifier == PropertyIdentifier.presentValue:
                            val = Base.extract_value_from_primitive_data(property_value)
                            self.point._trend(val)
                        elif property_identifier == PropertyIdentifier.statusFlags:
                            self.point.properties.status_flags = property_value
                        else:
                            self.point._log.warning(
                                f"Unsupported COV property identifier {property_identifier}"
                            )
                        if callable(self.callback):
                            self.point.log(
                                f"Calling callback for {self.point.properties.name}",
                                level="debug",
                            )
                            if asyncio.iscoroutinefunction(self.callback):
                                await self.callback(
                                    property_identifier=property_identifier,
                                    property_value=property_value,
                                )
                            elif hasattr(self.callback, "__call__"):
                                self.callback(
                                    property_identifier=property_identifier,
                                    property_value=property_value,
                                )
                            else:
                                self.point.log(
                                    f"Callback {self.callback} is not callable",
                                    level="error",
                                )
                await cov_fini_task_monitor
        except Exception as e:
            self.point.log(f"Error in COV subscription : {e}", level="error")

    def stop(self):
        self.point.log(
            f"Stopping COV subscription class for {self.point.properties.name}",
            level="debug",
        )
        self.cov_fini.set()
        self.point.cov_registered = False


class OfflineException(Exception):
    pass


def extract_value_from_primitive_data(value):
    if isinstance(value, float):
        return float(value)
    elif isinstance(value, Boolean):
        if value == int(1):
            return True
        else:
            return False
    elif isinstance(value, int):
        return int(value)
    elif isinstance(value, str):
        return str(value)
    else:
        return value



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#

# --- standard Python modules ---
# --- 3rd party modules ---

from collections import namedtuple
from typing import Any, Dict, List, Optional, Tuple, Union

from bacpypes3.primitivedata import Date, Time

# --- this application's modules ---
from ..utils.notes import note_and_log
from ..utils.lookfordependency import pandas_if_available

_PANDAS, pd, _, _ = pandas_if_available()


# ------------------------------------------------------------------------------

HistoryComponent = namedtuple("HistoryComponent", "index logdatum status choice")


class TrendLogProperties(object):
    """
    A container for trend properties
    """

    def __init__(self):
        self.device: Optional[Any] = None
        self.oid: Optional[Any] = None
        self.object_name: Optional[str] = None
        self.description: str = ""
        self.log_device_object_property: Optional[Any] = None
        self.buffer_size: int = 0
        self.record_count: int = 0
        self.total_record_count: int = 0
        self.log_interval: int = 0
        self.statusFlags: Optional[Any] = None
        self.status_flags: Dict[str, bool] = {
            "in_alarm": False,
            "fault": False,
            "overridden": False,
            "out_of_service": False,
        }
        self._history_components: List[HistoryComponent] = []
        self._df: Optional[pd.DataFrame] = None
        self.type: str = "TrendLog"
        self.units_state: str = "None"

    def __repr__(self):
        return "{} | Descr : {} | Record count : {}".format(
            self.object_name, self.description, self.record_count
        )

    @property
    def name(self) -> Optional[str]:
        return self.object_name


def TrendLog(*args: Tuple, **kwargs: Dict) -> "_TrendLog":
    trend = _TrendLog(*args, **kwargs)
    # update_properties_task = Task(trend.update_properties())
    # update_properties_task.start()
    # while not update_properties_task.done:
    #    pass
    # if "read_log_on_creation" in args:
    #    trend.read_log_buffer_task = Task(trend.read_log_buffer())

    return trend


@note_and_log
class _TrendLog(TrendLogProperties):
    """
    BAC0 simplification of TrendLog Object
    """

    def __init__(
        self,
        OID: Any,
        device: Optional[Any] = None,
        read_log_on_creation: bool = True,
        multiple_request: Optional[Any] = None,
    ):
        self.properties: TrendLogProperties = TrendLogProperties()
        self.properties.device = device
        self.properties.oid = OID
        self.update_properties_task: Optional[Any] = None
        self._last_index: int = 0
        if read_log_on_creation:
            self.read_log_buffer_task: Optional[Any] = None

    @staticmethod
    def read_logDatum(logDatum: Any) -> Tuple[str, Any]:
        for k, v in logDatum.__dict__.items():
            if v is None:
                continue
            else:
                return (k, v)

    async def update_properties(self) -> None:
        try:
            (
                self.properties.object_name,
                self.properties.description,
                self.properties.record_count,
                self.properties.buffer_size,
                self.properties.total_record_count,
                self.properties.statusFlags,
                self.properties.log_interval,
            ) = await self.properties.device.properties.network.readMultiple(
                "{addr} trendLog {oid} objectName description recordCount bufferSize totalRecordCount statusFlags logInterval".format(
                    addr=self.properties.device.properties.address,
                    oid=str(self.properties.oid),
                )
            )
            self.properties.description = str(self.properties.description)
        except Exception as error:
            raise Exception(f"Problem reading trendLog informations: {error}")

    async def _total_record_count(self) -> int:
        self.properties.total_record_count = (
            await self.properties.device.properties.network.read(
                "{addr} trendLog {oid} totalRecordCount".format(
                    addr=self.properties.device.properties.address,
                    oid=str(self.properties.oid),
                )
            )
        )
        return self.properties.total_record_count

    async def read_log_buffer(self) -> None:
        RECORDS = 10
        log_buffer = set()
        _actual_index = await self._total_record_count()
        start = max(_actual_index - self.properties.record_count + 1, self._last_index)
        _count = max(_actual_index - start, 0)
        steps = int(_count / RECORDS) + int(1 if (_count % RECORDS) > 0 else 0)

        self.log(f"Reading log : {start} {_count} {steps}", level="debug")

        _from = start
        for each in range(steps):
            range_params = ("s", _from, Date("1979-01-01"), Time("00:00"), RECORDS)
            _chunk = await self.properties.device.properties.network.readRange(
                "{} trendLog {} logBuffer".format(
                    self.properties.device.properties.address, str(self.properties.oid)
                ),
                range_params=range_params,
            )
            _from += len(_chunk)
            for chunk in _chunk:
                log_buffer.add(chunk)
        self._last_index = _from
        self.create_dataframe(log_buffer)

    def create_dataframe(self, log_buffer: set) -> None:
        for each in log_buffer:
            year, month, day, dow = each.timestamp.date
            year = year + 1900
            hours, minutes, seconds, ms = each.timestamp.time
            seconds = 0 if seconds == 255 else seconds
            ms = 0 if ms == 255 else ms
            _index = pd.to_datetime(
                f"{year}-{month}-{day} {hours}:{minutes}:{seconds}.{ms}",
                format="%Y-%m-%d %H:%M:%S.%f",
            )
            _choice, _logDatum = self.read_logDatum(each.logDatum)
            _status = each.statusFlags
            self._log.debug(f"{_index}, {_logDatum}, {_status}, {_choice}")
            his_component = HistoryComponent(_index, _logDatum, _status, _choice)
            if his_component not in self.properties._history_components:
                self.properties._history_components.append(his_component)

        if _PANDAS:
            df = pd.DataFrame(
                {
                    "index": [
                        each.index for each in self.properties._history_components
                    ],
                    self.properties.object_name: [
                        each.logdatum for each in self.properties._history_components
                    ],
                    "status": [
                        each.status for each in self.properties._history_components
                    ],
                    "choice": [
                        each.choice for each in self.properties._history_components
                    ],
                }
            )
            df = df.set_index("index")
            # df["choice"] = _choice
            # df[self.properties.object_name] = df['logDatum']

            self.properties._df = df
        else:
            # self.properties._history_components = (self.index, self.logdatum, self.status)
            self._log.warning(
                "Pandas not installed. Treating histories as simple list."
            )

    @property
    async def history(self) -> Union[Dict, Any]:
        await self.read_log_buffer()

        if not _PANDAS or self.properties._df is None:
            return dict(
                zip(
                    [each.index for each in self.properties._history_components],
                    [each.logDatum for each in self.properties._history_components],
                )
            )

        try:
            if not self.properties.log_device_object_property:
                self.properties.log_device_object_property = (
                    await self.properties.device.properties.network.read(
                        "{addr} trendLog {oid} logDeviceObjectProperty".format(
                            addr=self.properties.device.properties.address,
                            oid=str(self.properties.oid),
                        )
                    )
                )
            (
                objectType,
                objectAddress,
            ) = self.properties.log_device_object_property.objectIdentifier
            logged_point = self.properties.device.find_point(objectType, objectAddress)
        except (Exception, ValueError):
            logged_point = None

        serie = self.properties._df[self.properties.object_name].copy()
        serie.units = logged_point.properties.units_state if logged_point else "n/a"
        serie.name = ("{}/{}").format(
            self.properties.device.properties.name, self.properties.object_name
        )
        if not logged_point:
            serie.states = "unknown"
            serie.datatype = None
        else:
            if logged_point.properties.name in self.properties.device.binary_states:
                serie.states = "binary"
            elif logged_point.properties.name in self.properties.device.multi_states:
                serie.states = "multistates"
            else:
                serie.states = "analog"
            serie.datatype = objectType
        serie.description = self.properties.description

        return serie.sort_index()

    def chart(self, remove: bool = False) -> None:
        """
        Add point to the bacnet trending list
        """
        if not _PANDAS:
            self._log.error(
                "Pandas must be installed to use live chart feature. See documentation how how to run BAC0 in complete mode"
            )
        else:
            if remove:
                self.properties.device.properties.network.remove_trend(self)
            else:
                self.properties.device.properties.network.add_trend(self)

    def __repr__(self):
        return self.properties.__repr__()



import json
import os
from typing import Any, Dict, List, Optional, Set

from bacpypes3.app import Application
from bacpypes3.basetypes import BDTEntry, HostNPort

from ...core.utils.notes import note_and_log


@note_and_log
class BAC0Application:
    _learnedNetworks: Set = set()
    _cfg: Optional[Dict[str, Any]] = None

    def __init__(
        self, cfg: Dict[str, Any], addr: str, json_file: Optional[str] = None
    ) -> None:
        self._cfg = cfg
        self.cfg: Dict[str, Any] = self.update_config(cfg, json_file)
        self.localIPAddr: str = addr
        self.bdt: List[str] = self._cfg["BAC0"]["bdt"]
        self.device_cfg, self.networkport_cfg = self.cfg["application"]
        self.log(f"Configuration sent to build application : {self.cfg}", level="debug")

        self.app: Application = Application.from_json(self.cfg["application"])

    def register_as_foreign_device_to(self, host: str, lifetime: int = 900) -> None:
        np = self.app.get_object_name("NetworkPort-1")
        linklayer = self.app.link_layers[np.objectIdentifier]
        hnp = HostNPort(host)
        np.fdBBMDAddress = hnp
        linklayer.register(hnp.address, lifetime)

    def populate_bdt(self) -> None:
        np = self.app.get_object_name("NetworkPort-1")
        linklayer = self.app.link_layers[np.objectIdentifier]
        for addr in self.bdt:
            bdt_entry = BDTEntry(addr)
            np.bbmdBroadcastDistributionTable.append(bdt_entry)
            linklayer.add_peer(bdt_entry.address)

    def get_bacnet_ip_mode(self) -> str:
        return self.app.get_object_name("NetworkPort-1").bacnetIPMode

    def unregister_from_bbmd(self) -> None:
        self.app.unregister()

    def update_config(
        self, cfg: Dict[str, Any], json_file: Optional[str]
    ) -> Dict[str, Any]:
        if json_file is None:
            if os.path.exists(
                os.path.join(os.path.expanduser("~"), ".BAC0", "device.json")
            ):
                json_file = os.path.join(
                    os.path.expanduser("~"), ".BAC0", "device.json"
                )
                self.log("Using JSON Stored in user folder ~/.BAC0", level="info")

            else:
                json_file = os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "device.json"
                )
                self.log("Using default JSON configuration file", level="info")
        with open(json_file, "r") as file:
            base_cfg = json.load(file)

        base_cfg["application"][0].update(cfg["device"])
        base_cfg["application"][1].update(cfg["network-port"])
        return base_cfg



FILE_HEADER = """<?xml version="1.0"?>
<tags>
    <tag>
        <name>n/a</name>
        <group>n/a</group>
        <resourceLocator>
            <protocolName>n/a</protocolName>
            <object_type>n/a</object_type>
            <device_id>n/a</device_id>
            <data_type>n/a</data_type>
            <arraysize>n/a</arraysize>
            <conversion>n/a</conversion>
            <object_instance>n/a</object_instance>
            <object_property>n/a</object_property>
            <array_index>n/a</array_index>
            <write_priority>n/a</write_priority>
            <cov>n/a</cov>
        </resourceLocator>
        <encoding>n/a</encoding>
        <refreshTime>n/a</refreshTime>
        <accessMode>n/a</accessMode>
        <active>n/a</active>
        <TAGLOCATOR>n/a</TAGLOCATOR>
        <comment>n/a</comment>
        <simulator>
            <DataSimulator>n/a</DataSimulator>
            <Amplitude>n/a</Amplitude>
            <Simulator_offset>n/a</Simulator_offset>
            <Period>n/a</Period>
        </simulator>
        <scaling>
            <enableScaling>n/a</enableScaling>
            <scalingType>n/a</scalingType>
            <enableLimits>n/a</enableLimits>
            <factors>
                <s1>n/a</s1>
                <s2>n/a</s2>
                <s3>n/a</s3>
                <tagS1>n/a</tagS1>
                <tagS2>n/a</tagS2>
                <tagS3>n/a</tagS3>
            </factors>
            <limits>
                <eumin>n/a</eumin>
                <eumax>n/a</eumax>
                <elmin>n/a</elmin>
                <elmax>n/a</elmax>
            </limits>
        </scaling>
        <decimalDigits>
            <ddTag>n/a</ddTag>
            <ddDigits>n/a</ddDigits>
        </decimalDigits>
        <castType>n/a</castType>
        <default>n/a</default>
        <min>n/a</min>
        <max>n/a</max>
        <statesText>n/a</statesText>
    </tag>
"""

TAG = """    <tag>
        <name>{d[name]}</name>
        <group>{d[group]}</group>
        <resourceLocator>
            <protocolName>BACN</protocolName>
            <object_type>{d[object_type]}</object_type>
            <device_id>{d[device_id]}</device_id>
            <data_type>{d[data_type]}</data_type>
            <arraysize></arraysize>
            <conversion></conversion>
            <object_instance>{d[object_instance]}</object_instance>
            <object_property>{d[object_property]}</object_property>
            <array_index>{d[array_index]}</array_index>
            <write_priority>{d[write_priority]}</write_priority>
            <cov>{d[cov]}</cov>
        </resourceLocator>
        <encoding></encoding>
        <refreshTime>{d[refresh_time]}</refreshTime>
        <accessMode>{d[access_mode]}</accessMode>
        <active>{d[active]}</active>
        <TAGLOCATOR></TAGLOCATOR>
        <comment>{d[comment]}</comment>
        <simulator>
            <DataSimulator>Variables</DataSimulator>
            <Amplitude></Amplitude>
            <Simulator_offset></Simulator_offset>
            <Period></Period>
        </simulator>
        <scaling>
            <enableScaling>false</enableScaling>
            <scalingType>byFormula</scalingType>
            <enableLimits>false</enableLimits>
            <factors>
                <s1>1</s1>
                <s2>1</s2>
                <s3>0</s3>
                <tagS1></tagS1>
                <tagS2></tagS2>
                <tagS3></tagS3>
            </factors>
            <limits>
                <eumin>0</eumin>
                <eumax>100</eumax>
                <elmin></elmin>
                <elmax></elmax>
            </limits>
        </scaling>
        <decimalDigits>
            <ddTag></ddTag>
            <ddDigits></ddDigits>
        </decimalDigits>
        <castType></castType>
        <default></default>
        <min>-3.40282e+38</min>
        <max>3.40282e+38</max>
        <statesText></statesText>
    </tag>
"""

FILE_FOOTER = """</tags>
"""
